#!/usr/bin/python
#
# Use 'dig' command to get FQDN and address of local (or given) host name,
# as this is the only way of ensuring a timeout if there are
# network connectivity problems.
#
# Python bug: gethostbyname() can NEVER be interrupted, neither by signals,
# nor timer threads - it just hangs if no net connectivity. I've tried everything!
#
# Jason Vas Dias, Red Hat Inc., Oct 2004 
import os
import string
import time
import re
import commands
from   Hosts import Hosts

class Lookup:
    lookup_cmd=    '/usr/bin/dig +noall +nocmd +nocomments +answer +tries=2 +time=1 +search="`/bin/grep search /etc/resolv.conf | /bin/sed \'s/search //\'`"'
    hostname_cmd = '/bin/hostname'
    ip_cmd       = '/sbin/ip addr show scope global up | while read line; do if [[ $line = inet* ]]; then l=($line); a=${l[1]}; a=${a/\/*/}; b=${l[1]}; b=${b/*\//}; let n=${#l[*]}-1; echo $a $b ${l[$n]}; fi; done'
    ip4re=re.compile(r'^([0-9]{1,3})(((\.([0-9]){1,3}){1,3})?)$')
    ip6re=re.compile(r'^([0-9a-fA-F]{1,4}|:)?(:([0-9a-fA-F]{0,4})){0,7}$')

    def lookup_name(s, addr):        
        r = commands.getstatusoutput(s.lookup_cmd + ' -x ' + addr + ' PTR ' + addr + ' CNAME '  )
        if (r[0] == 0) and (len(r[1]) >0 ):
            cns=[]
            names=[]
            for a in r[1].split('\n'):
                (ipdn, ttl, cls, t, dn) = a.split()
                if t == 'CNAME':
                    cns.append(dn)
                else:
                    names.append(dn)
            return string.join(names,' ') + ' ' + string.join(cns,' ')
        else:
            s.loadHosts()
            if ( s.hosts.hosts.has_key(addr) ):
                i=0
                for i in range(0,len(s.hosts.hosts[addr])):
                    if type(s.hosts.hosts[addr][i])!=str:
                        s.hosts.hosts[addr].pop(i)
                return string.join(s.hosts.hosts[addr],' ')
        return None

    def lookup_addr(s, name):        
        r = commands.getstatusoutput(s.lookup_cmd+ ' ' + name + ' A ' + name + ' AAAA '+ name + ' CNAME ' )
        if (r[0] == 0) and (len(r[1])):
            cns=[]
            addrs=[]
            for a in r[1].split('\n'):
                (ipdn, ttl, cls, t, dn) = a.split()
                if t == 'CNAME':
                    cns.append(dn)
                else:
                    addrs.append(dn)
            return (ipdn, string.join(addrs,' ') + ' ' + string.join(cns,' '))
        else:
            s.loadHosts()
            if ( s.hosts.names.has_key(name) ):
                i=0
                for i in range(0,len(s.hosts.names[name])):
                    if type(s.hosts.names[name][i])!=str:
                        s.hosts.names[name].pop(i)
                debug("HOSTS addrs by name:", s.hosts.names[name])
                addr = s.hosts.names[name][0]
                if s.hosts.hosts.has_key(addr):
                    name = s.hosts.hosts[addr][0]
                debug("HOSTS addr by name:", name, addr)
                return (name, addr)
        return (None,None)

    def hostname(s):
        r = commands.getstatusoutput(s.hostname_cmd)
        if (r[0] != 0) or (len(r[1]) == 0):                
            return 'localhost'
        return r[1]
    
    def if_addrs(s):
        s.if_addr=[]
        # get list of active, global internet (ip4/ip6) addresses bound to local interfaces:
        r=commands.getstatusoutput(s.ip_cmd)
        if r[0] == 0:
            for l in r[1].split('\n'):
                addr,bs,ifn=l.split(' ')
                bits=32
                try:
                    bits=int(bs)
                except ValueError:
                    continue
                name = s.lookup_name(addr)                
                if name != None:
                    aliases=name.split()
                    if len(aliases) > 1:
                        s.if_addr.append([ addr, bits, ifn, aliases])
                    else:
                        s.if_addr.append([ addr, bits, ifn, name])
                else:
                    s.if_addr.append([ addr, bits, ifn ])
        if len(s.if_addr):
            for ifx in s.if_addr:
                if len(ifx) > 3:
                    return (ifx[3], ifx[0])
            return (None, s.if_addr[0][0])
        return (None, None)

    def loadHosts(s):
        st=os.stat("/etc/hosts")
        if st.st_mtime > s.hostsTime:
            s.hosts = Hosts("/etc/hosts",namesToo=True)

    def lookupLocalHost(s):
        (name, addr)=s.if_addrs()
        if( name == None ):
            s.name = 'localhost'
            s.addr = '127.0.0.1'
        elif type(name) == list:
            s.name = name[0]
            s.aliases=name[1:]                    
            s.addr = addr
        else:
            s.name = name
            s.addr = addr
        aliases = s.name.split()
        if len(aliases) > 1:
            s.name = aliases[0]
            s.aliases = aliases[1:]

    def __init__(s, host=None):
        s.aliases=[]
        s.name=None
        s.addr=None
        s.if_addr=[]
        s.hostsTime = 0
        s.loadHosts()
        isLocalHost=False
        if (host==None):
            isLocalHost=True
            host = s.hostname()
            debug( "Lookup: NONE:", host)
            if (host == 'localhost') or (host == 'localhost.localdomain') :
                s.lookupLocalHost()
                s.name=s.name.strip()
                s.addr=s.addr.strip()
                debug( 'Lookup:', s.name, s.aliases, s.addr )                    
                return
        debug( "Lookup: HOST:",host )
        if s.ip6re.match(host) or s.ip4re.match(host):
            s.name = s.lookup_name(host)                    
            if(s.name == None):
                (s.name,s.addr) = s.lookup_addr(host)
                if s.addr != None:
                    aliases = s.addr.split()
                    if len(aliases)>1:
                        s.addr = aliases[0]
                        s.aliases = aliases[1:]
            else:
                s.addr = host
                aliases = s.name.split()
                if len(aliases) > 1:
                    s.name = aliases[0]
                    s.aliases=aliases[1:]
        else:
            (name,s.addr) = s.lookup_addr(host)            
            debug( "Lookup: ", name, s.addr )
            if s.addr == None:
                s.name = None
            else:
                s.name = name
                aliases = s.addr.split()
                if len(aliases)>1:
                    s.addr = aliases[0]
                    s.aliases = aliases[1:]
        if isLocalHost and ( ( s.name == None ) or ( s.addr == None )  ):
            s.lookupLocalHost()
        if s.name != None:
            s.name=s.name.strip()
        if s.addr != None:
            s.addr=s.addr.strip()
        debug( 'Lookup:', s.name, s.aliases,  s.addr)

localhost = Lookup()

if __name__=='__main__':
    import sys
    if len( sys.argv ) >= 2:
        l = Lookup(sys.argv[1])
    else:
        l = Lookup()
    print 'NAME:',l.name, 'ADDR:', l.addr, 'ALIASES:', l.aliases
