#
# BIND Resource Record Field Edit Widgets
#
# Jason Vas Dias, Red Hat Inc., October 2004

import os
import stat
import copy
import string
import socket
import gtk
import gtk.glade
import gnome.ui
import gnome
import gobject

import RR
import DNS

RDNC_KEY_NAME=_('Remote Nameserver Control Security Key')

class Widget:
    def __init__(s, d, r, a, k):
        s.dialog = d
        s.rr = r
        s.attr = a
        s.key = k
        s.gui=d.gui
        if( a != None):
            s.data = str(a)
        else:
            s.data = None        

    def value(s):
        return s.entry.get_text()

    def save(s):        
        v = s.value()
        if v != s.data:
            s.data = v
            setattr(s.rr,s.key,v)
            return True
        return False
    
    def widget(s):
        s.entry = gtk.Entry()
        s.entry.set_text(str(s.data))
        s.entry.show()
        return s.entry
    

class DNS_Name(Widget):
    def __init__(s, d, r, a, k):
        Widget.__init__(s, d, r, a, k)
        debug( 'DNS_Name NEW: ',a.__class__)
        debug( '   ATTR:',a)
        if k == 'name' and (a == None):
            s.data=r.Name()
        else:
            s.data=str(a)        

    def select_origin(s,w):
        n = s.originSelect.get_active()
        if (n > 0) and (n < len(s.origins)):
            v = s.value()
            o = s.origins[n-1].origin_label()
            if len(o) and (o[-1] != '.'):
                o+='.'
            e = s.entry.get_text()
            if len(e) > 0 :
                a = DNS.split_dots(e)
                e = a[0] + '.' + o
            else:
                e = o
            s.entry.set_text(e)        
    
    def widget(s):
        if isinstance(s.rr, RR.SOA) and (s.key == 'name'):
            s.entry = gtk.Label(s.data)
            return s.entry
        else:
            s.entry = Widget.widget(s)        
            s.entry.connect('focus_out_event', s.focus_out)
            if (isinstance(s.rr,RR.PTR)):
                s.originSelect = gtk.combo_box_new_text()
                view=None
                if s.dialog.bind.isViewed:
                    view=s.rr.zone.nmdcKey[1]
                s.origins = s.dialog.bind.zone_origins(ztype=DNS.FORWARD,view=view)
                if len(s.origins):
                    s.vb = gtk.VBox()
                    s.vb.pack_start(s.entry)
                    s.originSelect.append_text(_('Select Origin'))
                    s.originSelect.set_active(0)
                    for zone in s.origins:
                        s.originSelect.append_text(zone.origin_label())
                    s.originSelect.connect('changed',s.select_origin)
                    s.vb.pack_start(s.originSelect)
                return s.vb
        return s.entry

    def answer(s,response,arg):
        if ( response == gtk.RESPONSE_OK ):
            s.entry.set_text(arg + '.')
        s.focus_out(s.entry, None, isAnswer=True)

    def focus_out(s, entry, event, isAnswer = False, dialogIssued=None):
        v = entry.get_text()
        if (len(v)==0) or ( (v[-1] != '*') and ((v[-1] != '.' ) or ((len(v)>=2) and (v[-2] == '\\')))):
            vp=v
            if ( len(s.rr.zone.zone_origin) ):
                if (  s.rr.zone.zone_origin != '.'):
                    v += '.' + s.rr.zone.zone_origin
                else:
                    v += s.rr.zone.zone_origin
            else:
                s.gui.warning(s.dialog.dialog,_("DNS Name Warning"),_("An absolute DNS name is required here. '.' appended."))
                s.entry.set_text(v + '.')
                s.entry.grab_focus()
                if( type(dialogIssued) == dict ):
                    dialogIssued['yes']=1
                return False
            if hasattr(s.rr,'zone') and  ( s.rr.zone.type < 0 ) and not isAnswer:
                ans = s.gui.question(s.dialog.dialog, _("DNS Name Question"), _("Do you really want to use this domain name:\n%s\nor should I append a '.'?") % v)
                ans.question(s.answer,vp)
                if( type(dialogIssued) == dict ):
                    dialogIssued['yes']=1
                return False
            entry.set_text(v)
        if(s.key == 'name') and not s.rr.zone.In(v):
            s.gui.modalError(s.dialog.dialog,_("DNS Name Error"),_("Name %s is out-of-zone %s.") %(v, s.rr.zone.zone_origin))
            entry.set_text(s.rr.zone.zone_origin)
            entry.grab_focus()
            if( type(dialogIssued) == dict ):
                    dialogIssued['yes']=1
            return False
        errors = s.attr.check(v) 
        if (errors != None) and len(errors):
            for code,error in errors:
                if code == -1:
                    s.gui.modalError(s.dialog.dialog,_("DNS Name Error"), _('Bad DNS Name: %s -\n%s.') %(v, error))
                    entry.grab_focus()
                    if( type(dialogIssued) == dict ):
                        dialogIssued['yes']=1
                else:
                    s.gui.modalError(s.dialog.dialog,_("DNS Name Warning"), _('Poor DNS Name: %s -\n%s.') %(v, error),warning=True)
                    if( type(dialogIssued) == dict ):
                        dialogIssued['yes']=0
        return False

    def value(s):
        n = s.entry.get_text()
        if hasattr(s.attr,'zone_origin'):
            if ( n == s.attr.zone_origin ):
                return '@'
            elif (len(n) >= len(s.attr.zone_origin)) and (n[-len(s.attr.zone_origin):].upper() == s.attr.zone_origin.upper()):
                return n[:-(len(s.attr.zone_origin)+1)]
            else:
                return n
        return n
    
    def save(s):
        n = s.value()
        if (n == None) or (len(n) <= 0):
            return _('DNS Name cannot be empty.')
        origin=''
        if hasattr(s.attr,'zone_origin'):
            origin=s.attr.zone_origin
        if hasattr(s.rr,'not_origin') and len(n) and (n in ['@', origin]):
            return _('The Zone Origin DNS name cannot be specified here.')
        debug( 'DNS.Name SAVE:', n,  s.key, origin)
        if issubclass(s.rr.__class__, RR.RR):
            if( s.key == 'name' ):
                if( s.rr.name == None):
                    name = s.rr.Name()
                    if ( n != name.value() ):
                        s.rr.name = DNS.Name(n,s.attr.zone)
                        debug( 'DNS.Name NEW name',n)
                        return True
        if s.attr.data != n:
            s.attr.data = n
            debug( 'DNS.Name UPDATE', n, s.attr)
            return True
        return False
                           

class DNS_TTL(Widget):
    def __init__(s, d, r, a, k):
        Widget.__init__(s, d, r, a, k)
    
    def D_change(s, adj, p=None):
        d=adj.get_value()
        if (d >= 7):
            adj.set_value(0.0)
            w=s.weeks.get_value()
            s.weeks.set_value(w+1.0)

    def H_change(s, adj, p=None):
        h=adj.get_value()
        if (h >= 24):
            adj.set_value(0.0)
            d=s.days.get_value()
            s.days.set_value(d+1.0)

    def M_change(s, adj, p=None):
        m=adj.get_value()
        if (m >= 60):
            adj.set_value(0.0)
            h=s.hours.get_value()
            s.hours.set_value(h+1.0)

    def S_change(s, adj, p=None):
        sec=adj.get_value()
        if (sec >= 60):
            adj.set_value(0.0)
            m=s.minutes.get_value()
            s.minutes.set_value(m+1.0)
 
    def widget(s):
        if (s.attr == None) and hasattr(s.rr,'TTL'):
            s.attr=s.rr.TTL()
        if (s.attr == None):
            s.attr=DNS.TTL('0')
        t = s.attr.wdhms()
        hb = gtk.HBox()
        s.weeks = gtk.Adjustment(lower=0, upper=3550, step_incr=1, page_incr=1, page_size=1)
        s.days  = gtk.Adjustment(lower=0, upper=   7, step_incr=1, page_incr=1, page_size=1)
        s.hours = gtk.Adjustment(lower=0, upper=  24, step_incr=1, page_incr=1, page_size=1)
        s.minutes=gtk.Adjustment(lower=0, upper=  60, step_incr=1, page_incr=1, page_size=1)
        s.seconds=gtk.Adjustment(lower=0, upper=  60, step_incr=1, page_incr=1, page_size=1)
        s.days.connect('value_changed',s.D_change)
        s.hours.connect('value_changed',s.H_change)
        s.minutes.connect('value_changed',s.M_change)
        s.seconds.connect('value_changed',s.S_change)
        s.wsb   = gtk.SpinButton(s.weeks)
        s.wsb.set_numeric(True)
        s.wsb.set_max_length(4)
        s.wsb.set_value(float(t['W']))
        s.dsb   = gtk.SpinButton(s.days)
        s.dsb.set_numeric(True)
        s.dsb.set_max_length(1)
        s.dsb.set_value(float(t['D']))
        s.hsb   = gtk.SpinButton(s.hours)
        s.hsb.set_numeric(True)
        s.hsb.set_max_length(2)
        s.hsb.set_value(float(t['H']))
        s.msb   = gtk.SpinButton(s.minutes)
        s.msb.set_numeric(True)
        s.msb.set_max_length(2)
        s.msb.set_value(float(t['M']))        
        s.ssb   = gtk.SpinButton(s.seconds)
        s.ssb.set_numeric(True)
        s.ssb.set_max_length(2)
        s.ssb.set_value(float(t['S']))
        vb = gtk.VBox()
        vb.pack_start(gtk.Label(_('weeks')),False)
        vb.pack_start(s.wsb,False)
        hb.pack_start(vb,False)
        vb = gtk.VBox()
        vb.pack_start(gtk.Label(_('days')),False)
        vb.pack_start(s.dsb,False)
        hb.pack_start(vb,False)
        vb = gtk.VBox()
        vb.pack_start(gtk.Label(_('hours')),False)
        vb.pack_start(s.hsb,False)
        hb.pack_start(vb,False)
        vb = gtk.VBox()
        vb.pack_start(gtk.Label(_('minutes')),False)
        vb.pack_start(s.msb,False)
        hb.pack_start(vb,False)
        vb = gtk.VBox()
        vb.pack_start(gtk.Label(_('seconds')),False)
        vb.pack_start(s.ssb,False)
        hb.pack_start(vb,False)
        hb.show_all()
        return hb
    
    def value(s):
        t=''
        if( s.weeks.get_value() > 0.0 ):
            t=str(int(s.weeks.get_value()))+'W'
        if( s.days.get_value() > 0.0 ):
            t+=str(int(s.days.get_value()))+'D'
        if( s.hours.get_value() > 0.0 ):
            t+=str(int(s.hours.get_value()))+'H'
        if( s.minutes.get_value() > 0.0 ):
            t+=str(int(s.minutes.get_value()))+'M'
        if( s.seconds.get_value() > 0.0 ):
            t+=str(int(s.seconds.get_value()))+'S'
        return t
    
    def save(s):
        inherited_ttl=None
        v = s.value()
        vt=DNS.TTL(v)
        if issubclass(s.rr.__class__,RR.RR) and (s.key == 'ttl') and (s.rr.ttl == None):
            inherited_ttl = s.rr.TTL()
        if( inherited_ttl == None ) or (vt.seconds != s.attr.seconds):
            if( inherited_ttl != None ):
                s.rr.ttl = DNS.TTL(v)
                debug( 'NEW TTL:',s.rr.ttl)
                return True
            elif(vt.seconds != s.attr.seconds):
                debug( 'UPDATE TTL:',vt.seconds,s.attr.seconds,v)
                s.attr.input( v )
                return True
        return False
    
class IPV4_Address(Widget):
    def __init__(s, d, r, a, k):
        Widget.__init__(s, d, r, a, k)

    def prefix(s):
        if hasattr(s,'pfx'):
            pfx=int(s.pfx.get_value())
        else:
            pfx=0
        if( s.attr.prefix != None ) and (pfx > 0) and (pfx < 32):
            return '/'+str(pfx)
        return ''
        
    def prefix_change(s,pfx_adj):
        pfx = int(pfx_adj.get_value())
        bits=0
        for o in s.octsb:
            if bits >= pfx: 
                o.set_sensitive(False)
            else:
                o.set_sensitive(True)
            bits += 8
    
    def more_octs(s,b):
        s.hb.remove(s.moreOctsButton)
        s.octs.append('0')
        adj = gtk.Adjustment(lower=0, upper= 255, step_incr=1, page_incr=1, page_size=1)
        sb=gtk.SpinButton(adj)
        sb.set_numeric(True)
        sb.set_value(float(0))
        s.octsb.append(sb)
        s.octadj.append(adj)
        s.hb.pack_start(sb)
        if hasattr(s,'is_origin'):
            max_octs=3
        else:
            max_octs=4
        if len(s.octs) < max_octs:            
            s.hb.pack_start(gtk.Label('.'))
            s.moreOctsButton=gtk.Button(stock=gtk.STOCK_ADD,label='')
            s.moreOctsButton.connect('clicked',s.more_octs)
            s.hb.pack_start(s.moreOctsButton,False)
        s.hb.show_all()

    def select_origin(s,w):
        n = s.originSelect.get_active()
        if (n > 0) and ((n-1) < len(s.origins)):
            v = s.value()
            o = s.origins[n-1].origin_label()
            octs = o.split('.')
            i=0
            for oct in octs:
                s.octsb[i].set_value(float(oct))
                i+=1
            if( i < len(s.octsb)):
                s.octsb[i].grab_focus()
        
    def widget(s):
        prefix=32
        if s.attr.prefix != None:
            prefix = s.attr.prefix
        debug( 'IPV4 Address: ', str(s.attr), 'PREFIX:', prefix, 'is_origin:', hasattr(s,'is_origin'), 'origin_only:', hasattr(s,'origin_only'))
        s.octs = s.attr.octs()
        if not (hasattr(s,'is_origin') or hasattr(s,'origin_only')):
            while len(s.octs) < 4:
                s.octs.append('0')                
        s.octsb=[]
        s.octadj=[]
        is_A_rr=False
        if (s.rr.__class__ == RR.A) and ( s.key == 'ADDRESS' ):
            debug( 'IPv4 A RECORD')
            s.vb=gtk.VBox()
            is_A_rr=True
        s.hb = gtk.HBox()
        bits=0
        total_bits = len(s.octs) * 4
        for oct in s.octs:
            adj = gtk.Adjustment(lower=0, upper= 255, step_incr=1, page_incr=1, page_size=1)
            sb=gtk.SpinButton(adjustment=adj)
            sb.set_max_length(3)
            sb.set_numeric(True)
            sb.set_value(float(oct))
            s.octsb.append(sb)
            s.octadj.append(adj)
            bits += 4
            if ( bits >  prefix ):
                sb.set_sensitive(False)
            s.hb.pack_start(sb,False)
            if bits < total_bits:
                s.hb.pack_start(gtk.Label('.'),False)
        if( len(s.octs) < 4) and (not hasattr(s,'origin_only')):
            debug( 'IPV4 MORE OCTS')
            s.hb.pack_start(gtk.Label('.'))
            s.moreOctsButton = gtk.Button(stock=gtk.STOCK_ADD,label='')
            s.moreOctsButton.connect('clicked',s.more_octs)
            s.hb.pack_start(s.moreOctsButton)
        if s.attr.prefix != None :
            s.hb.pack_start(gtk.Label('/'))
            s.pfx=gtk.Adjustment(lower=0, upper= 32, step_incr=1, page_incr=1, page_size=1)
            s.pfx.connect('value_changed',s.prefix_change)                
            s.pfxsb=gtk.SpinButton(s.pfx)
            s.pfxsb.set_numeric(True)
            s.pfxsb.set_value(prefix)
            s.hb.pack_start(s.pfxsb)
        if is_A_rr :
            s.originSelect = gtk.combo_box_new_text()
            view=None
            if s.dialog.bind.isViewed:
                view=s.rr.zone.nmdcKey[1].strip('"')
            s.origins = s.dialog.bind.zone_origins(ztype=DNS.IPV4.REVERSE,view=view)
            if len(s.origins):
                s.vb.pack_start(s.hb)
                s.originSelect.append_text(_('Select IPv4 Prefix'))
                s.originSelect.set_active(0)
                for zone in s.origins:
                    s.originSelect.append_text(zone.origin_label())
                s.originSelect.connect('changed',s.select_origin)
                s.vb.pack_start(s.originSelect)
                return s.vb
            else:
                return s.hb
        else:
            return s.hb
        
    def value(s):
        ip = ''
        n_o=0
        for o in s.octsb:
            ip += str(int(o.get_value()))
            n_o+=1
            if n_o < len(s.octsb):
                ip+='.'
        ip += s.prefix()
        return ip
    
    def save(s):
        ip = s.value()
        if ip != s.attr.value():
            s.attr.input(s.value(), prefixAllowed=(len(s.prefix())>0))
            return True
        return False

class IPV4_ReverseAddress(IPV4_Address):
    def __init__(s, d, r, a, k):
        Widget.__init__(s, d, r, a, k)
        s.r_addr = s.attr
        s.attr = s.r_addr.address()

    def value(s):
        ip = IPV4_Address.value(s)
        return s.r_addr.reverse(ip)+'.in-addr.arpa.'         

    def widget(s):
        if( issubclass(s.rr.__class__,RR.RR)):
            if( s.key == 'name') and (s.rr.__class__ in [ RR.SOA, RR.NS ]):
                debug( 'IPV4 REVERSE ORIGIN ONLY')
                s.origin_only=True
        oocts=s.attr.origin_address.octs()
        debug( 'LEN ORIGIN OCTS:', len(oocts))
        if( len(oocts) == 0):
            s.is_origin = True
        w = IPV4_Address.widget(s)
        for i in range(0,len(oocts)):
            s.octsb[i].set_sensitive(False)
        return w
    
    def save(s):
        v = s.value()
        if( hasattr(s.r_addr,'zone_origin') ):
            if( v.upper() == s.r_addr.zone_origin.upper()):
                v = '@'
            elif (len(v) > len(s.r_addr.zone_origin)) and (v[-len(s.r_addr.zone_origin):].upper() ==  s.r_addr.zone_origin.upper()):
                v = v[:-(len(s.r_addr.zone_origin)+1)]            
        if( issubclass(s.rr.__class__,RR.RR)):
            if( s.key == 'name'):
                if( s.rr.name == None):
                    name = s.rr.Name()
                    if( v != name.value() ):
                        s.rr.name = DNS.IPV4.ReverseAddress(v,s.rr.zone)
                        s.r_addr = s.rr.name
                        s.attr = s.r_addr.address()
                        return True
        if( s.r_addr.addr != v):
            s.r_addr.addr = v
            s.attr = s.r_addr.address()
            return True
        return False

class IPV6_Address(Widget):
    def __init__(s, d, r, a, k):
        Widget.__init__(s, d, r, a, k)

    def check_hex(s, entry, event):
        hex = entry.get_text()
        bad = False
        for c in hex:
            if not c in '0123456789ABCDEFabcdef':
                bad = True
        if bad:
            s.dialog.gui.modalError(s.dialog.dialog,_('Invalid IPv6 Address "%s" contains invalid hex digits : only [0-9] [A-F] [a-f] allowed.') % hex)
            entry.grab_focus()
        l = entry.get_max_length()
        if(len(hex) < l ) and not( (entry==s.entries[-1]) and (hasattr(s,'is_origin') or hasattr(s,'origin_only'))):
            debug( 'FILL:',len(hex),l)
            for i in range(len(hex),l):
                hex = '0' + hex
                entry.set_text(hex)
        
    def prefix_change(s, pfx):
        prefix = int(pfx.get_value())
        bits = 0
        for e in s.entries:
            bits+=16
            if( bits > prefix ):
                e.set_editable(False)
                e.set_sensitive(False)

    def more_hex(s,b):
        s.hb.remove(s.moreHexButton)
        nibs = (len(s.entries) + 1) * 4
        entry = gtk.Entry(max=4)
        entry.set_text('0000')
        entry.set_max_length(4)
        entry.set_width_chars(4)
        entry.connect('focus_out_event',s.check_hex)
        if( len(s.entries) > 1):
            v=s.entries[-1].get_text()
            l=s.entries[-1].get_max_length()
            while len(v) < l:
                v='0'+v
            s.entries[-1].set_text(v)
        s.entries.append(entry)
        s.hb.pack_start(entry,False)
        s.hb.pack_start(gtk.Label(':'),False)        
        if( nibs <= 28 ):
            s.moreHexButton=gtk.Button(stock=gtk.STOCK_ADD)
            s.moreHexButton.connect('clicked',s.more_hex)
            s.hb.pack_start(s.moreHexButton,False)
        s.hb.show_all()
        
    def select_origin(s,w):
        n = s.originSelect.get_active()
        if ( n > 0 ) and ( (n-1) < len(s.origins)):
            i=0
            n=0
            for h in s.origins[n-1].SOA.Name().address.split_hex():
                s.entries[i].set_text(h)
                if ( h != '0000' ):
                    n = i+1
                i+=1
            if( n < len(s.entries)):
                s.entries[n].grab_focus()
        
    def widget(s):
        s.hb = gtk.HBox()
        sens = 0
        a = s.attr
        s.entries=[]
        subHex=0
        if hasattr(a,'sensitivity'):            
            sens = a.sensitivity
            if ((sens % 4) != 0):
                subHex=sens
        nibs=0
        addr=None
        if ( hasattr(a.address,'address') ):
            if a.address.address != None:
                a_len = len(a.address.address)
            else:
                a_len = 0
            addr=a.address
        else:
            a_len = 32
            addr=a
        debug( 'IPV6 Widget: ',addr.address, 'IS ORIGIN:',hasattr(s,'is_origin'), 'ORIGIN ONLY:', hasattr(s,'origin_only'))
        if( hasattr(s,'is_origin') and ((addr.address==None) or (len(addr.address)==0) )):
            hexes = [ '0000' ]
            addr.address = '0000'
            a_len = 4
        else:
            debug( 'IPV6 addr:',addr)
            hexes = addr.split_hex()
            debug( 'IPV6 HEXES:', hexes)
            debug( 'IPV6 A_LEN: ',a_len, 'SENSITIVITY:',sens)
            if (not hasattr(s,'is_origin')) and (not hasattr(s,'origin_only')):
                while len(hexes) < 8:
                    hexes.append('0000')
                    a_len+=4
            else:
                while (len(hexes)*4) > a_len:
                    hexes.pop(-1)
                if a_len > (len(hexes)*4):
                    subHex=a_len
                    hexes.append(addr.address[-(a_len-(len(hexes)*4)):])
            if( (len(hexes) == 0) and (len(addr.address)>0) and(len(addr.address)<4)):
                hexes=[addr.address]
                if(sens < 4):
                    sens=4
            debug( 'LEN HEXES:', len(hexes),a_len,hexes)
        for hex in hexes:
            debug( 'HEX:',hex,subHex)
            if (subHex > 0) and (subHex < 4) :
                s.dnPfx=gtk.Entry(max=subHex)
                s.dnPfx.set_max_length(subHex)
                s.dnPfx.set_width_chars(subHex)
                s.dnPfx.set_text(hex[:subHex])
                s.dnPfx.set_editable(False)
                s.dnPfx.set_sensitive(False)
                s.hb.pack_start(s.dnPfx,False)
                s.entries.append(s.dnPfx)
                if hasattr(s,'origin_only'):
                    break
                hex=hex[subHex:]
                entry = gtk.Entry(max=4-subHex)
                entry.set_text(hex)
                entry.set_max_length(4-subHex)
                entry.set_width_chars(4-subHex)
                s.dnPfx=entry
                nibs+=subHex
            else:
                entry = gtk.Entry(max=4)
                entry.set_text(hex)
                entry.set_max_length(4)
                entry.set_width_chars(4)
                entry.connect('focus_out_event',s.check_hex)            
            entry.connect('focus_out_event',s.check_hex)
            digits=4
            if (subHex > 0):
                if subHex  < 4:
                    digits=4-subHex
                subHex -= 4
            nibs+=digits
            if( nibs <= sens ) and not(hasattr(s,'is_origin')):
                entry.set_editable(False)
                entry.set_sensitive(False)
            s.hb.pack_start(entry,False)
            if( nibs < 32 ):
                s.hb.pack_start(gtk.Label(":"),False)
            s.entries.append(entry)
        if( (nibs < 32) and not(hasattr(s,'origin_only'))):
            s.moreHexButton=gtk.Button(stock=gtk.STOCK_ADD)
            s.moreHexButton.connect('clicked',s.more_hex)
            s.hb.pack_start(s.moreHexButton,False)            
        if addr.prefix != None:
            s.hb.pack_start(gtk.Label('/'))
            s.pfx=gtk.Adjustment(lower=0, upper= 128, step_incr=1, page_incr=1, page_size=1)
            s.pfx.connect('value_changed',s.prefix_change)                
            s.pfxsb=gtk.SpinButton(s.pfx)
            s.pfxsb.set_value(float(addr.prefix))
            s.hb.pack_start(s.pfxsb)
            s.prefix_change(s.pfx)
        if s.rr.__class__ == RR.AAAA:
            s.vb=gtk.VBox()
            s.originSelect = gtk.combo_box_new_text()
            view=None
            if s.dialog.bind.isViewed:
                view=s.rr.zone.nmdcKey[1]
            s.origins = s.dialog.bind.zone_origins(ztype=DNS.IPV6.REVERSE_ARPA,view=view)
            if len(s.origins):
                s.vb.pack_start(s.hb)
                s.originSelect.append_text(_('Select IPv6 Prefix'))
                s.originSelect.set_active(0)
                for zone in s.origins:
                    s.originSelect.append_text(zone.origin_label())
                s.originSelect.connect('changed',s.select_origin)
                s.vb.pack_start(s.originSelect)
                return s.vb
            else:
                return s.hb
        return s.hb

    def value(s):
        ip=''
        for entry in s.entries:
            nib=entry.get_text()
            if (len(nib)==4) or (hasattr(s.dnPfx) and (entry == s.dnPfx)):
                ip += nib + ':'
            else:
                ip += nib
        if len(ip) and (ip[-1] == ':'):
            ip = ip[:-1]
        if ( hasattr(s,'pfx') ):
            ip += '/' + str(int(s.pfx.get_value()))
        return ip

    def save(s):
        v=DNS.IPV6.Address(s.value(),DNS.IPV6.HEX_FORMAT,hasattr(s,'pfx'))
        debug( 'IPV6 SAVE: VALUE',v)
        if( v.address != s.attr.address ):
            s.attr.address=v.address
            s.attr.prefix = v.prefix
            return True
        return False

class IPV6_AAAA(IPV6_Address):
    def __init__(s, d, r, a, k):
        IPV6_Address.__init__(s, d, r, a, k)
    
    def save(s):
        v=DNS.IPV6.Address(s.value(),DNS.IPV6.HEX_FORMAT)
        if v.address != s.attr.address.address:
            s.attr.address.address=v.address
            return True
        return False        

class IPV6_A6(IPV6_Address):
     def __init__(s, d, r, a, k):
        IPV6_Address.__init__(s, d, r, a, k)

     def prefix_change(s,w):
         pfx = int(s.pfx.get_value())
         i=0
         debug( '    PFX CHANGE:',pfx, hasattr(s,'hb'))
         for bits in range(int(pfx/16)*16,128,16):
             if i < len(s.entries):
                 w=s.entries[i]
                 w.set_sensitive(True)             
             else:
                 entry = gtk.Entry(max=4)
                 entry.set_text('0000')
                 entry.set_max_length(4)
                 entry.set_width_chars(4)
                 entry.connect('focus_out_event',s.check_hex)
                 if len(s.entries) > 0:
                     s.hb.pack_start(gtk.Label(':'),False)
                 s.entries.append(entry)
                 s.hb.pack_start(entry,False)
                 if i > 0:
                     for j in range (i,0,-1):                     
                         s.entries[j].set_text( s.entries[ j - 1 ].get_text() )                         
                     s.entries[0].set_text('0000')
             i+=1
         n_pfx_hexes = int(pfx/16)
         n_hexes = 8 - n_pfx_hexes
         debug( '  n entries: ',len(s.entries),'n_hexes:',n_hexes)
         cs=s.hb.get_children()                  
         for e in range(len(s.entries),n_hexes, -1):
             w=s.entries.pop(0)
             s.hb.remove(w)
             cs.pop(0)
             if len(cs)> 0:
                 s.hb.remove(cs.pop(0))             
         if pfx == 0:
             s.nameWidget.entry.set_sensitive(False)
         else:
             s.nameWidget.entry.set_sensitive(True)
         s.hb.show_all()

     def value(s):
         ip=''
         pfx=s.pfx.get_value()
         for entry in s.entries:
            nib=entry.get_text()
            ip += nib + ':'
         ip = ip[:-1]
         if pfx > 0:
             ip = '::' + ip
         elif len(s.entries) < 8:
             ip += '::'
         return ip

     def save(s):
         r=False
         v=s.value()
         pfx = int(s.pfx.get_value())
         debug( 'IPV6 A6 SAVE - addr: ',v, 'PFX: ',pfx, 'PFX_N:', s.nameWidget.value())
         if (pfx <= 0):
             addr=DNS.IPV6.Address(v,DNS.IPV6.HEX_FORMAT)
             if (addr.address==None) or (len(addr.address) < 32):
                 return _('An 128 bit IPv6 address is required.')
         else:
             if s.pfx == 128:
                 if len(v):
                     return _('An IPv6 address cannot be specified with an 128 bit prefix.')                 
             n=s.nameWidget.value()
             if len(n) <= 0:
                 return _('The DNS name of the prefix must be entered.')
             if n in [ '@', s.nameWidget.attr.zone_origin ]:
                 return _('You cannot specify the Zone Origin as the prefix name.')             
         if pfx != s.attr.prefix:
             s.attr.prefix = pfx
             r=True
         if (str(s.attr.address) != v):
             s.attr.address.input(v,DNS.IPV6.HEX_FORMAT)
             if s.attr.address.address == None:
                 return _('Invalid IPv6 Address:\n%s\n.') % str(s.attr.address.errors)
             debug( '  A6 ADDR B4 PFX: ', s.attr.address.address)
             s.attr.address.address = '%0.32lx' % ( s.attr.address.int() & long((2**(128-pfx))-1) )
             r=True
         if str(s.attr.prefix_name) != s.nameWidget.value():
             s.attr.prefix_name.data =  s.nameWidget.value()
             r=True
         return r
         
     def widget(s):
         debug( 'IPV6 A6 Widget', s.attr.prefix, s.attr.address, s.attr.prefix_name, hasattr(s.rr,'new_record'))
         s.entries=[]
         s.hb=None
         s.vb=gtk.VBox()
         s.is_origin = True
         s.origin_only=True
         if( hasattr(s.rr,'new_record') and (s.rr.index==-1) ):
             s.attr.address.address=''
             s.attr.prefix_name.data=''
         s.pfxFrame = gtk.Frame(label=_('Prefix Bits Length'))
         s.pfx=gtk.Adjustment(lower=0, upper= 128, step_incr=1, page_incr=16, page_size=1)
         s.pfx.connect('value_changed',s.prefix_change)                
         s.pfxsb=gtk.SpinButton(s.pfx)
         s.pfxsb.set_value(float(s.attr.prefix))
         s.pfxFrame.add(s.pfxsb)
         s.vb.pack_start(s.pfxFrame)
         s.addrFrame=gtk.Frame(label=_('Address Suffix'))         
         s.addrWidget=IPV6_Address.widget(s)
         if( hasattr(s.rr,'new_record') and (s.rr.index==-1) ):
             cs=s.hb.get_children()
             s.hb.remove(cs.pop(-1))             
         s.addrFrame.add(s.addrWidget)
         s.vb.pack_start(s.addrFrame)
         s.nameFrame=gtk.Frame(label=_('Prefix DNS Name'))
         s.nameWidget = DNS_Name(s.dialog, s.rr, s.attr.prefix_name, ['ADDRESS','prefix_name'])
         s.nameW = s.nameWidget.widget()
         s.nameFrame.add(s.nameW)
         s.vb.pack_start(s.nameFrame)
         s.prefix_change(s.pfx)
         return s.vb
         
class IPV6_ReverseARPAddress(IPV6_Address):
    def __init__(s, d, r, a, k):
        IPV6_Address.__init__(s, d, r, a, k)
        debug( 'IPV6 REVERSE WIDGET:', str(s.attr.address), str(s.attr.origin_address))
        if (s.attr.origin_address.address == None) or (len(s.attr.origin_address.address) == 0):
            s.is_origin = True
            debug( 'IS ORIGIN')
        else:
            debug( 'IPV6 SENSITIVITY:', len(s.attr.origin_address.address))
            s.attr.sensitivity = len(s.attr.origin_address.address)
            if (s.rr.__class__ in [ RR.SOA, RR.NS ]) and (s.key == 'name'):
                s.origin_only=True
                debug( 'ORIGIN ONLY', hasattr(s,'origin_only'))

    def value(s):
        a=''
        for e in s.entries:
            v=e.get_text()
            l=e.get_max_length()
            if not(hasattr(s,'is_origin') or hasattr(s,'origin_only')):
                while(len(v)<l):
                    v='0'+v
            a+=v        
        return DNS.IPV6.reverse(string.join(a[:],'.'))+'.ip6.arpa.'
    
    def save(s):
        v = s.value()
        debug( 'REVERSE IPV6 SAVE: value:',v,'origin:',s.attr.zone_origin)
        if( v == str(s.attr.address)):
            return FALSE
        if(s.attr.zone_origin != None) and (len(v) >= len(s.attr.zone_origin)):
            if v == s.attr.zone_origin:
                v='@'
            elif v[-len(s.attr.zone_origin):] == s.attr.zone_origin:
                v= v[:-(len(s.attr.zone_origin)+1)]
        if( issubclass(s.rr.__class__,RR.RR)):
            if( s.key == 'name'):
                if( s.rr.name == None):
                    name = s.rr.Name()
                    if( v != name.addr ):                        
                        s.rr.name = DNS.IPV6.ReverseARPAddress(v,s.rr.zone)
                        s.attr = s.rr.name
                        return True
                    else:
                        return FALSE
        a=DNS.IPV6.ReverseARPAddress(v,s.rr.zone)
        s.attr.addr=v
        s.attr.origin = a.origin
        s.attr.zone_origin = a.zone_origin
        s.attr.address = a.address
        if (s.attr.origin_address == None) or (s.attr.origin_address.address == None):
            s.attr.origin_address = a.address
            s.attr.origin = v
            s.attr.zone_origin=v
            s.attr.addr='@'
        debug( 'REVERSE IPV6 SAVE:', s.attr.addr, str(s.attr), str(a))
        return True

class IPV6_ReverseINTAddress(IPV6_ReverseARPAddress):
    def __init__(s, d, r, a, k):
        IPV6_ReverseARPAddress.__init__(s, d, r, a, k)

class DNS_Address(Widget):
    def __init__(s, d, r, a, k):
        Widget.__init__(s, d, r, a, k)

    def type_changed(s,c):
        changed=False
        if (s.combo.get_active() == 0) and (s.attr.address.__class__ != DNS.IPV4.Address):
            s.attr.address=DNS.IPV4.Address('0.0.0.0')
            s.addressW=IPV4_Address(s.dialog,s.rr,s.attr.address,'address')
            changed=True
        elif (s.combo.get_active() == 1) and (s.attr.address.__class__ != DNS.IPV6.Address):
            s.attr.address=DNS.IPV6.Address('::')
            s.addressW=IPV6_Address(s.dialog,s.rr,s.attr.address,'address')
            changed=True
        if changed:
            s.hb.remove(s.addressWidget)
            s.addressWidget=s.addressW.widget()
            s.hb.pack_start(s.addressWidget,False)        
            s.hb.show_all()
            
    def widget(s):
        s.hb = gtk.HBox()
        s.combo = gtk.combo_box_new_text()
        s.combo.append_text(_('IPv4 Address'))
        s.combo.append_text(_('IPV6 Address'))
        s.combo.set_active(0)
        s.combo.connect('changed',s.type_changed)
        s.hb.pack_start(s.combo,False)
        if( hasattr(s.attr,'address') ):
            if s.attr.address.__class__ == DNS.IPV4.Address:
                s.addressW = IPV4_Address(s.dialog, s.rr, s.attr.address, 'address')
            elif s.attr.address.__class__ == DNS.IPV6.Address:
                s.addressW = IPV6_Address(s.dialog, s.rr, s.attr.address, 'address')                
            elif s.attr.address.__class__ == str:
                if s.attr.address.find(':') > 0:
                    s.addressW = IPV6_Address(s.dialog, s.rr, s.attr.address, 'address')
                else:
                    s.addressW = IPV4_Address(s.dialog, s.rr, s.attr.address, 'address')
        else:
            s.attr.address=DNS.IPV4.Address('0.0.0.0')
            s.addressW=IPV4_Address(s.dialog,s.rr,s.attr.address,'address')
        s.addressWidget = s.addressW.widget()
        s.hb.pack_start(s.addressWidget,False)
        return s.hb

    def __str__(s):
        return str(s.attr.address)

    def value(s):
        return s.addressW.value()

    def save(s):
        v = s.value()
        if v != str(s.attr.address):
            s.attr.address = s.attr.address.__class__(v)
            return True
        return False

class NSAP_Address(Widget):
    def __init__(s,d,r,a,k):
        Widget.__init__(s,d,r,a,k)

    def check_hex(s,w,e,problem=None):
        a = s.entry.get_text()
        if (len(a)>=2) and (a[:2] == '0x'):
            a=a[2:]
        aok=''
        badChars = ''
        for c in a:
            if c not in '.0123456789ABCDEFabcdef':
                badChars += c
            else:
                aok += c
        if len(badChars):
            if (problem != None) and (type(problem) == dict):
                problem['problem']=_('Illegal characters in NSAP address:\n %s \nremoved.' % badChars)
                s.entry.set_text(aok)
                s.entry.grab_focus()
                return False
            s.gui.modalError(s.dialog.dialog,_('NSAP Address Error'),
                             _('Illegal characters in NSAP address:\n %s \nremoved.' % badChars)
                            )
            s.entry.set_text(aok)
            s.entry.grab_focus()
        return False
    
    def widget(s):
        s.hb = gtk.HBox()
        s.entry = gtk.Entry()
        a = s.attr.label()
        if (len(a)>=2) and (a[:2] == '0x'):
            a=a[2:]
        s.entry.set_width_chars(max(len(a),20))
        s.hb.pack_start(gtk.Label('0x'),False)
        s.entry.set_text(a)
        s.entry.connect('focus_out_event',s.check_hex)
        s.hb.pack_start(s.entry,False)
        return s.hb

    def value(s):
        return '0x' + s.entry.get_text()

    def save(s):
        v = s.value()
        if v != s.attr.label():
            s.attr.address = s.attr.strip(v)
            return True
        return False

class NSAP_ReverseAddress(NSAP_Address):
    def check_hex(s,w,e,problem=None):
        debug( 'NSAP_ReverseAddress Check' )
        NSAP_Address.check_hex(s,w,e,problem)
        if (problem != None) and (type(problem)==dict) and problem.has_key('problem'):
            return False
        v = s.value()
        debug( 'VALUE:',v)
        rv = s.attr.strip(v)
        debug( 'NSAP_ReverseAddress Check',rv, s.attr.origin_address.address)
        if ((s.attr.origin_address.address != None) and (len(s.attr.origin_address.address)>0) and
            (  ( len(rv) < len(s.attr.origin_address.address) )
             or( rv[:len( s.attr.origin_address.address ) ] != s.attr.origin_address.address)
            )) :
            if (problem != None) and (type(problem)==dict):
                problem['problem'] = _('NSAP Address\n"%s"\nis out-of-zone\n"%s".') % (v,s.attr.origin_address.label())
            else:
                s.gui.modalError(s.dialog.dialog,_('Invalid Reverse NSAP Address'),
                                 _('NSAP Address\n"%s"\nis out-of-zone\n"%s".') % (v,s.attr.origin_address.label())
                                )
            s.entry.set_text(s.attr.label()[2:])
            s.entry.grab_focus()
        return False
        
    def save(s):        
        v = s.value()
        debug( 'NSAP_ReverseAddress SAVE:',v)
        bad={}
        if v != s.attr.label():
            s.check_hex(s.entry,None,bad)
            if bad.has_key('problem'):
                return bad['problem']
            rv = s.attr.strip(v)
            s.attr.address = rv
            if (s.attr.origin_address.address != None) and (len(s.attr.origin_address.address)>0):
                rv = rv[len( s.attr.origin_address.address ):]
            ra = ''            
            for i in range( len(rv)-1,-1,-1):
                c = rv[i]
                ra+=c+'.'            
            ra=ra[:-1]
            s.attr.raddr = ra
            if (s.attr.origin_address.address == None) or (len(s.attr.origin_address.address)==0):
                s.attr.raddr += '.nsap.int.'
                if (s.attr.zone_origin == None) or (len(s.attr.zone_origin)==0):
                    s.attr.zone_origin = s.attr.raddr
                debug( 'NSAP NULL ORIGIN - zone_origin:',s.attr.zone_origin)
                s.attr.origin_address = DNS.NSAP.Address(s.attr.reverse(s.attr.zone_origin))
            debug( 'NSAP_ReverseAddress SAVED:', ra, rv)
            return True
        else:
            return False

    def widget(s):
        e = NSAP_Address.widget(s)
        if ( isinstance(s.rr, RR.SOA) or isinstance(s.rr,RR.NS) ) and (
            s.key == 'name'
           ):
            s.entry.set_sensitive(False)
        return e

class DNS_Email(DNS_Name):
    def __init__(s, d, r, a, k):
        DNS_Name.__init__(s, d, r, a, k)
        s.data=a.emailLabel()
        
    def focus_out(s, entry, event, isAnswer = False, dialogIssued=None):
        s.entry.set_text(s.attr.dnsLabel(s.entry.get_text()))
        DNS_Name.focus_out(s,entry,event,isAnswer,dialogIssued)
        s.data=s.attr.emailLabel(s.entry.get_text())
        s.entry.set_text(s.data)

    def save(s):        
        v = s.attr.dnsLabel(s.entry.get_text())
        if( s.attr.data != v ):
            if issubclass(s.rr.__class__, RR.RR) and len(s.rr.zone.zone_origin):
                if( len(s.attr.data) > len(s.rr.zone.zone_origin) ):
                    s.attr.data=s.attr.data[:-(len(s.rr.zone.zone_origin)+1)]
                    return True
            s.attr.data=v
            return True
        return False


class DNS_NoneAddress(Widget):
    def __init__(s, d, r, a, k):
        Widget.__init__(s, d, r, a, k)

    def widget(s):
        return gtk.Label('none')

    def value(s):
        return 'none'
    
class DNS_AnyAddress(Widget):
    def __init__(s, d, r, a, k):
        Widget.__init__(s, d, r, a, k)

    def widget(s):
        return gtk.Label('any')

    def value(s):
        return 'any'
    
class DNS_LocalAddress(Widget):
    def __init__(s, d, r, a, k):
        Widget.__init__(s, d, r, a, k)

    def widget(s):
        return gtk.Label('localhost')

    def value(s):
        return 'localhost'
    
class DNS_LocalNetsAddress(Widget):
    def __init__(s, d, r, a, k):
        Widget.__init__(s, d, r, a, k)

    def widget(s):
        return gtk.Label('localnets')

    def value(s):
        return 'localnets'        

class DNS_ACL_List(Widget):
    def __init__(s, d, r, a, k):
        Widget.__init__(s, d, r, a, k)

class DNS_Key_List(Widget):
    def __init__(s, d, r, a, k):
        Widget.__init__(s, d, r, a, k)

class DNS_ACL(Widget):
    class Selection:
        acl=None
        
        def __init__(s,parentACL=None):
            s.parentACL=parentACL
            
        def selected(s,tree):
            selection = tree.get_selection()
            if( selection != None ):
                model, selected = selection.get_selected()
                if (model != None) and  (selected != None):
                    return  model.get_value(selected, 1)
            return None

        def select(s, t, p1=0,p2=0):
            s.acl = s.selected(s.aclTree)
            debug( 'ACL SELECTED:',s.acl            )
            return False
                
        def widget(s):
            s.acl=None
            s.aclList = gtk.ListStore( gobject.TYPE_STRING, gobject.TYPE_PYOBJECT )
            s.aclTree = gtk.TreeView( model=s.aclList )
            s.aclTree.set_headers_visible(False)
            s.aclTree.connect('row_activated',s.select)
            s.aclTree.connect('button_release_event',s.select)
            s.aclCell = gtk.CellRendererText()
            s.aclColm = gtk.TreeViewColumn( None, s.aclCell, text=0)
            s.aclTree.append_column(s.aclColm)
            n_acls=0
            for acl in DNS.ACL.namedACLs.keys():
                debug("NAMED ACL SELECTION:", acl, s.parentACL.contains_acl(acl))
                if (acl != s.parentACL.name) and not(s.parentACL.contains_acl(acl)):                    
                    s.aclList.append([acl, DNS.ACL.namedACLs[acl]])
                    n_acls += 1
            if n_acls > 0:
                s.hb=gtk.HBox()
                s.aclSW=gtk.ScrolledWindow()
                s.aclSW.set_policy(gtk.POLICY_AUTOMATIC,gtk.POLICY_AUTOMATIC)
                s.aclSW.add_with_viewport(s.aclTree)
                s.hb.pack_start(s.aclSW)
                return s.hb
            else:
                s.hb=gtk.HBox()
                s.hb.pack_start(gtk.Label(_('There are no named ACLs that\ncan be added to this ACL.')))
                s.no_selection = True
                return s.hb
    
    class KeySelection:
        acl=None
        
        def __init__(s,parentACL=None):
            s.parentACL=parentACL
            
        def selected(s,tree):
            selection = tree.get_selection()
            if( selection != None ):
                model, selected = selection.get_selected()
                if (model != None) and  (selected != None):
                    return  model.get_value(selected, 1)
            return None
        
        def select(s, t, p1=0,p2=0):
            s.key = s.selected(s.keyTree)
            debug( 'KEY SELECTED:',s.key            )
            return False
                
        def widget(s):
            s.key=None
            s.keyList = gtk.ListStore( gobject.TYPE_STRING, gobject.TYPE_PYOBJECT )
            s.keyTree = gtk.TreeView( model=s.keyList )
            s.keyTree.set_headers_visible(False)
            s.keyTree.connect('row_activated',s.select)
            s.keyTree.connect('button_release_event',s.select)
            s.keyCell = gtk.CellRendererText()
            s.keyColm = gtk.TreeViewColumn( None, s.keyCell, text=0)
            s.keyTree.append_column(s.keyColm)
            n_keys=0
            for key in DNS.Key.namedKeys.keys():
                if not(s.parentACL.contains_key(key)):                    
                    s.keyList.append([key, DNS.Key.namedKeys[key]])
                    n_keys += 1
            if n_keys > 0:                    
                s.keySW=gtk.ScrolledWindow()
                s.keySW.set_policy(gtk.POLICY_AUTOMATIC,gtk.POLICY_AUTOMATIC)
                s.keySW.add_with_viewport(s.keyTree)
                return s.keySW
            else:
                s.hb=gtk.HBox()
                s.hb.pack_start(gtk.Label(_('There are no named Keys that\ncan be added to this ACL.')))
                s.no_selection = True
                return s.hb
    
    class SaveAs:
        def name_check(s):
            s.name = None
            name=s.entry.get_text()
            if (len(name)==0):
                s.aclGUI.dialog.bind.gui.modalError(s.aclGUI.dialog.bind.gui.gui,_("ACL Name Error"),
                                             _('You must enter a non-empty ACL name.')
                                            )
                return False
            if (DNS.ACL.namedACLs.has_key(name)):
                s.aclGUI.dialog.bind.gui.modalError(s.aclGUI.dialog.bind.gui.gui,_("ACL Name Error"),
                                             _('An ACL with this name already exists.')
                                            )
                return False
            if s.aclGUI.dialog.bind.keys.has_key(name):
                s.aclGUI.dialog.bind.gui.modalError(s.aclGUI.dialog.bind.gui.gui,_("ACL Name Error"),
                                             _('A security key with this name already exists.')
                                            )
                return False
            if name in [ 'none', 'any', 'localhost', 'localnets' ]:
                s.aclGUI.dialog.bind.gui.modalError(s.aclGUI.dialog.bind.gui.gui,_("ACL Name Error"),
                                             _('The ACL name cannot be an ACL keyword.')
                                            )
                return False
            s.name = name
            return True
        
        def acl_name(s,w,e):
            if not s.name_check():
                w.grab_focus()
            return False

        def ok(s,b):
            if not s.name_check():
                s.entry.grab_focus()
                return False            
            acl = 'acl ' + s.name + s.aclGUI.attr.out() + "\n"
            s.aclGUI.attr.name = s.name
            s.aclGUI.dialog.bind.nmdc[['acl',s.name]]=acl
            s.aclGUI.dialog.bind.nmdc.update()
            s.aclGUI.dialog.bind.acls[s.name]=s.aclGUI.attr
            DNS.ACL.namedACLs[s.name]=s.aclGUI.attr
            s.aclGUI.dialog.bind.zoneTree.load()
#            s.aclGUI.aclList.clear()
#            s.aclGUI.attr.addrs=[ DNS.ACL.namedACLs[s.name] ]
            s.aclGUI.aclTree.set_cursor((0,))
            s.aclGUI.acl_element_change(s.aclGUI.aclTree,1)
            s.aclGUI.saveACLbutton.hide()
            s.aclGUI.aclVB.remove(s.aclGUI.saveACLbutton)
            if hasattr(s.aclGUI,'aclEditDialog'):
                s.aclGUI.aclEditDialog.set_title("ACL " + s.name)
            return False
        
        def __init__(self,s):
            self.aclGUI = s
            if( len(s.attr.addrs) == 0 ):
                return
            if(s.editContainer != None):
                s.editContainer.hide()
                s.editHB.remove(s.editContainer)
                s.editContainer.destroy()
            s.editContainer = gtk.VBox()
            self.entry = gtk.Entry()
            self.entry.connect('focus_out_event',self.acl_name)            
            s.aclFrame=gtk.Frame(_('Enter Access Control List Name'))
            s.aclFrame.add(self.entry)
            s.editContainer.pack_start(s.aclFrame)
            self.okButton = gtk.Button(stock=gtk.STOCK_OK)
            self.okButton.connect('clicked',self.ok)
            s.editContainer.pack_start(self.okButton,False)
            s.editHB.pack_start(s.editContainer)
            s.editHB.show_all()
    
    def __init__(s, d, r, a, k):
        Widget.__init__(s, d, r, a, k)
        if hasattr(d,'editDialog'):
            s.parent = d.editDialog
        elif hasattr(d,'gui'):
            s.parent = d.gui.gui
        else:
            s.parent =None
            debug( 'ACL WIDGET: UNKNOWN PARENT')
        debug( 'ACL WIDGET NEW:',d,r,a,k)

    def selected(s,tree):
        selection = tree.get_selection()
        if( selection != None ):
            model, selected = selection.get_selected()
            if (model != None) and  (selected != None):
                return (selected,) + model.get(selected, 0, 1)
        return (None,None,None)

    def acl_element_ok(s, b, ele):
        element =  s.aclList.get_value(ele,1)
        if( element.__class__ != DNS.ACL):
            s.editWidget.save()
            value = s.editWidget.attr
        else:
            value=element
        no=''
        if hasattr(value,'notted') and value.notted:
            no = '!'
        s.aclList.set_value(ele,0,no+str(value))
        s.aclList.set_value(ele,1,value)
        return False
    
    def acl_element_remove(s, b, ele):
        s.aclList.remove(ele)
        s.acl_element_change(s.aclTree)
        return False

    def n_acls(s):
        return s.aclList.iter_n_children(None)-1
    
    def acl_element_add(s,b,a=None):
        debug( 'ACL ADD:',s.editWidget        )
        if  isinstance(s.editWidget,Widget):
            debug( 'ACL ADD Widget')
            s.editWidget.save()
            attr = s.editWidget.attr
        elif isinstance(s.editWidget,DNS_ACL.Selection):
            debug( 'ACL ADD SELECTION')
            if( s.editWidget.acl == None ):
                return False            
            attr = s.editWidget.acl
        elif isinstance(s.editWidget,DNS_ACL.KeySelection):
            debug( 'ACL ADD KEY SELECTION')
            if( s.editWidget.key == None ):
                return False            
            attr = s.editWidget.key
        elif( hasattr(s,'aclWidget') ):
            debug( 'ACL ADD aclWidget'            )
            attr = s.aclWidget.attr
        elif ( a != None):
            debug( 'ACL ADD a')
            attr = a
        else:
            return False
        no=''
        if hasattr(attr,'notted') and attr.notted:
            no='!'
        debug( 'ACL ADD:', attr.__class__,str(attr))
        s.aclList.append([no + str(attr),attr])
        n_acls=s.n_acls()
        s.aclTree.set_cursor((n_acls,))
        s.acl_element_activated(s.aclTree,(n_acls,),1)
        return False      

    def acl_toggle_not(s, b, a):
        a.notted = not ( hasattr(a,'notted') and (a.notted == True) )
        if a.notted:
            b.set_label("!")
        else:
            b.set_label(" ")
            
    def editElement(s, a, iter=None):
        debug( 'ACL EDIT ELEMENT:', a, iter)
        s.editContainer = gtk.VBox()
        s.eHB = gtk.HBox()
        s.notButton = gtk.ToggleButton()
        if hasattr(a,'notted') and a.notted:
            s.notButton.set_active(True)
            s.notButton.set_label("!")
        else:
            s.notButton.set_active(False)
            s.notButton.set_label(" ")
        s.notButton.connect("toggled",s.acl_toggle_not,a)
        s.eHB.pack_start(s.notButton,False,False)
        if a.__class__  == DNS.ACL :
            s.editWidget = gtk.Button("Edit")
            debug( 'EDIT ACL',iter)
            s.editWidget.connect('clicked', s.editACL, a, iter)
            s.eHB.pack_start(s.editWidget,False,False)
        elif a.__class__ == DNS_ACL.Selection:
            s.editWidget  = a
            w=s.editWidget.widget()
            if hasattr(a,'no_selection'):
                s.eHB.remove(s.notButton)
            s.eHB.pack_start(w)
        elif a.__class__ == DNS_ACL.KeySelection:
            s.editWidget  = a
            w=s.editWidget.widget()
            if hasattr(a,'no_selection'):
                s.eHB.remove(s.notButton)
            s.eHB.pack_start(w)
        else:
            s.editWidget = s.dialog.gui.widget( a )(s.dialog, s.rr, a, s.key)
            s.eHB.pack_start( s.editWidget.widget(),False )
        debug( 'ACL WIDGET:',s.editWidget)
        s.editContainer.pack_start( s.eHB,False )
        s.adhb = gtk.HBox()
        if not(( a.__class__ in [ DNS_ACL.Selection, DNS_ACL.KeySelection ] ) and hasattr(a,'no_selection')):
            if iter != None:
                s.okButton = gtk.Button(stock=gtk.STOCK_OK)
                s.okButton.connect('clicked',s.acl_element_ok,iter)
                s.deleteButton = gtk.Button(stock=gtk.STOCK_REMOVE)
                s.deleteButton.connect('clicked',s.acl_element_remove,iter)
                s.adhb.pack_start(s.deleteButton,False)
            else:
                s.okButton = gtk.Button(stock=gtk.STOCK_ADD)
                s.okButton.connect('clicked',s.acl_element_add,a)
            s.adhb.pack_start(s.okButton,True,False)
            s.editContainer.pack_start(s.adhb,False)
        return s.editContainer

    def acl_element_change(s,tree,new=None):
        iter,ads,a=s.selected(tree)
        if s.editContainer != None:
            s.editContainer.hide()
            s.editHB.remove(s.editContainer)
            s.editContainer.destroy()
            s.editContainer = None
        if( new == None):
            iter = None
            attr=copy.deepcopy(a)
        else:
            attr=a
        if( a != None):
            debug( 'ACL ELEMENT CHANGE:', attr)
            s.editContainer = s.editElement(attr,iter)
        if ( s.editContainer != None ) and ( s.editContainer.__class__ != DNS_ACL ):            
            s.editHB.pack_start(s.editContainer)
        s.editHB.show_all()
        
    def acl_element_activated(s, tree, path, column):
        s.acl_element_change(tree,path)
        return False
    
    def new_acl_element_activated(s, tree, path, column):
        s.acl_element_change(tree)
        return False

    def acl_element_selected(s, tree, event):
        s.acl_element_change(tree,event)
        return False

    def new_acl_element_selected(s, tree, event):
        s.acl_element_change(tree)
        return False
    
    def load(s):
        s.aclList.clear()
        firstEle=None
        path=0
        if s.attr.addrs != None:
            for acl in s.attr.addrs:
                debug( 'LOAD ACL ELE:', acl )
                notted=''
                if hasattr(acl,'notted') and acl.notted:
                    notted='!'
                ele = s.aclList.append([ notted + str(acl), acl ])
                if firstEle == None:
                    firstEle = (path,)
                path+=1
        s.newList.clear()
        newEle=s.newList.append([_('IPv4 Address'),   DNS.IPV4.Address('0.0.0.0/32',True,True)])
        s.newList.append([_('IPv6 Address'),   DNS.IPV6.Address('::/128',DNS.IPV6.HEX_FORMAT,True)])
        s.newList.append([_('None'),           DNS.NoneAddress()])
        s.newList.append([_('Any'),            DNS.AnyAddress()])
        s.newList.append([_('Local Host'),     DNS.LocalAddress()])
        s.newList.append([_('Local Networks'), DNS.LocalNetsAddress()])
        s.newList.append([_('Named Key'),      DNS_ACL.KeySelection(s.attr)])
        s.newList.append([_('Named ACL'),      DNS_ACL.Selection(s.attr)])
        s.newList.append([_('ACL'),            DNS.ACL(None,s.dialog.bind.nmdc)])
        values=None
        if( type(s.rr) == dict ) and s.rr.has_key('values'):
            values = s.rr['values']
        elif (str(s.rr.__class__)=='Options.Option') and (s.rr.option.has_key('values')):
            values = s.rr.option['values']
        if values != None:
            ele=newEle
            while ele != None:                
                if s.newList.get_value(ele,0) not in values:
                    nextEle=s.newList.iter_next(ele)
                    if( ele == newEle):                        
                        newEle = nextEle                        
                    s.newList.remove(ele)
                    ele=nextEle
                else:
                    ele=s.newList.iter_next(ele)
        if( firstEle != None ):
            s.aclTree.set_cursor(firstEle)
            s.acl_element_change(s.aclTree,firstEle)
        else:
            s.newTree.set_cursor((0,))
            s.acl_element_change(s.newTree)
            
    def acl_elements_value(s, model, path, iter, user_data=None):
        ele=str(model.get_value(iter, 1)) + '; '
        debug( 'ACL VALUE ELEMENT:',ele)
        s.acl_elements_string +=  ele
        return False                            
    
    def value(s):
        s.acl_elements_string = ''
        s.aclList.foreach(s.acl_elements_value)
        debug( 'ACL VALUE:',s.acl_elements_string)
        return s.acl_elements_string

    def save_acl_element(s, model, path, iter, user_data=None):
        ele = model.get_value(iter, 1)
        s.attr.addrs.append( ele )
        debug( 'ACL ELEMENT SAVED:', str(ele))
        return False                            
        
    def save(s):
        debug( 'ACL SAVE ',s.rr.__class__, s.attr.name, s.attr.key)
        s.attr.addrs=[]
        s.aclList.foreach(s.save_acl_element)
        if (s.attr.name != None):
            if( type(s.attr.name) != list ):
                name = [ 'acl', s.attr.name ]
            else:
                name = s.attr.name[:]
            fullName = name[:]
            aclName = name[-1]
            if s.attr.nmdc.__contains__(name):                
                del s.attr.nmdc[name]
            else:
                name=name[:-1]
                if(  s.attr.nmdc.__contains__(name) ):
                    name += [ aclName ]
            debug( 'SAVING ACL:', name)
            s.attr.nmdc[name] = 'acl '+aclName+' '+s.attr.out()+'\n'            
            s.attr.nmdc.update()            
            debug( 'ACL SAVE: saved bind.acls[ ',aclName,']' )
            s.dialog.bind.acls[aclName]=s.attr
            DNS.ACL.namedACLs[aclName]=s.attr
            s.dialog.bind.zoneTree.load()
        elif s.attr.key != None:
            debug( 'SAVING un-named ACL:', s.attr.key)
            if s.attr.nmdc.__contains__( s.attr.key ):
                del s.attr.nmdc[ s.attr.key  ]

            if ( 'listen-on' in s.attr.key ) or ( 'listen-on-v6' in s.attr.key ):
                if ( len(s.attr.key) == 3 ):
                    s.attr.nmdc[ s.attr.key  ] = s.attr.key[-2] + ' ' + s.attr.key[-1] + ' ' + s.attr.out() + '\n'
                else:
                    s.attr.nmdc[ s.attr.key  ] = s.attr.key[-1] + ' ' + s.attr.out() + '\n'
            else:
                s.attr.nmdc[ s.attr.key  ] = s.attr.key[-1] + ' ' + s.attr.out() + '\n'

            s.attr.nmdc.update()
        
    def aclDialogResponse(s,d,r):
        if r == gtk.RESPONSE_OK:
            s.aclWidget.save()
            if hasattr(s,'aclList'):
                if s.aclIter == None:
                    s.acl_element_add(s,1)
                else:
                    s.aclList.set_value(s.aclIter,0,str(s.aclList.get_value(s.aclIter,1)))
        s.editDialog.destroy()
        return False
            
    def aclDialog(s, acl,iter=None):
        s.aclIter=iter
        debug( 'ACL DIALOG: ',iter)
        s.editDialog = gtk.Dialog(parent=s.parent, title=str(acl), buttons=
                                  (gtk.STOCK_OK,     gtk.RESPONSE_OK,
                                   gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL
                                  )
                                 )
        s.editDialog.connect('response',s.aclDialogResponse)
        s.aclWidget = DNS_ACL( s.dialog, s.rr, acl, acl.name )
        s.aclWidget.aclEditDialog = s.editDialog
        s.editDialog.vbox.pack_start(s.aclWidget.widget())
        s.editDialog.show_all()
        s.aclEdited = acl

    def editACL(s,b,acl,iter=None):
        s.aclDialog(acl,iter)
        
    def save_as_acl(s,b):
        debug( 'SAVE AS ACL')
        if( s.n_acls() != 0):
            s.save()
            DNS_ACL.SaveAs(s)
    
    def widget(s):
        s.hb = gtk.HBox()
        s.aclList = gtk.ListStore( gobject.TYPE_STRING, gobject.TYPE_PYOBJECT )
        s.newList = gtk.ListStore( gobject.TYPE_STRING, gobject.TYPE_PYOBJECT )
        s.aclTree = gtk.TreeView( model=s.aclList )
        s.newTree = gtk.TreeView( model=s.newList )
        s.aclTree.set_headers_visible(False)
        s.newTree.set_headers_visible(False)
        s.aclSW =   gtk.ScrolledWindow()
        s.aclSW.set_policy(gtk.POLICY_AUTOMATIC,gtk.POLICY_AUTOMATIC)
        s.newSW =   gtk.ScrolledWindow()
        s.newSW.set_policy(gtk.POLICY_AUTOMATIC,gtk.POLICY_AUTOMATIC)
        s.aclTree.connect('row_activated',s.acl_element_activated)
        s.newTree.connect('row_activated',s.new_acl_element_activated)
        s.aclTree.connect('button_release_event',s.acl_element_selected)
        s.newTree.connect('button_release_event',s.new_acl_element_selected) 
        s.aclCell = gtk.CellRendererText()
        s.newCell = gtk.CellRendererText()
        s.aclColm = gtk.TreeViewColumn( None, s.aclCell, text=0 )
        s.newColm = gtk.TreeViewColumn( None, s.newCell, text=0 )
        s.aclTree.append_column(s.aclColm)
        s.newTree.append_column(s.newColm)        
        s.aclFrame = gtk.Frame(label=_('Address Match List'))
        s.aclSW.add_with_viewport(s.aclTree)
        s.aclVB=gtk.VBox()
        s.aclVB.pack_start(s.aclSW)
        if s.attr.name == None:
            s.saveACLbutton = gtk.Button("Save As ACL")
            s.saveACLbutton.connect("clicked",s.save_as_acl)
            s.aclVB.pack_start(s.saveACLbutton,False)
        s.aclFrame.add(s.aclVB)
        s.editFrame = gtk.Frame(label=_('Edit List Element'))        
        s.editHB = gtk.HBox()
        s.editContainer= gtk.VBox()
        s.editWidget = gtk.Label('')        
        s.editContainer.pack_start(s.editWidget)
        s.editHB.pack_start(s.editContainer)
        s.editFrame.add(s.editHB)
        s.newFrame = gtk.Frame(label=_('New List Element'))
        s.newSW.add_with_viewport(s.newTree)
        s.newFrame.add(s.newSW)
        s.hb.pack_start(s.aclFrame)
        s.hb.pack_start(s.editFrame)
        s.hb.pack_start(s.newFrame)
        s.load()
        s.hb.show_all()
        return s.hb

class DNS_SortList(Widget):
    def selected(s,tree):
        selection = tree.get_selection()
        if( selection != None ):
            model, selected = selection.get_selected()
            if (model != None) and  (selected != None):
                return (selected,) + model.get(selected, 1)
        return (None,None)

    def addSource(s,b):
        s.currentACL.addrs=[ DNS.ACL(None,s.currentACL.nmdc), s.currentACL.addrs[0] ]
        for c in s.sourceHB.get_children():
            s.sourceHB.remove(c)
        s.sourceACL = DNS_ACL(s.dialog,s.aclValues,s.currentACL.addrs[0],None)
        s.sourceHB.pack_start(s.sourceACL.widget())
        s.preferLabel.set_text(_('RR Set Order Address Match List'))
        
    def editOrder(s,iter,acl):
        debug('SORTLIST editOrder:',acl.__class__)
        if s.currentACL != None:
            if s.sourceACL != None:
                s.sourceACL.save()
            if s.preferACL != None:
                s.preferACL.save()
        s.currentACL = acl
        s.currentIter= iter
        for c in s.sourceHB.get_children():
            s.sourceHB.remove(c)
        for c in s.preferHB.get_children():
            s.preferHB.remove(c)
        if len(acl.addrs)==0:
            acl.addrs.append(DNS.ACL({},s.attr.nmdc))
        if len(acl.addrs)>1:
            debug('SORTLIST editOrder:',acl.addrs[0].__class__, acl.addrs[1].__class__)
            s.sourceACL = DNS_ACL(s.dialog,s.aclValues,acl.addrs[0],None)
            s.preferACL = DNS_ACL(s.dialog,s.aclValues,acl.addrs[1],None)
            s.preferLabel.set_text(_('RR Set Order Address Match List'))
        else:
            debug('SORTLIST editOrder:',acl.addrs[0].__class__)
            s.sourceACL = None
            s.preferACL = DNS_ACL(s.dialog,s.aclValues,acl.addrs[0],None)
            s.preferLabel.set_text(_('Source and RR Set Order Address Match List'))
        if s.sourceACL != None:
            s.sourceHB.pack_start(s.sourceACL.widget())
        else:
            s.addSourceButton = gtk.Button(stock=gtk.STOCK_ADD)
            s.addSourceButton.connect('clicked',s.addSource)
            vb=gtk.VBox()
            vb.pack_start(s.addSourceButton,False)
            s.sourceHB.pack_start(vb,False)
        debug('preferACL:',s.preferACL.attr.__class__)
        s.preferHB.pack_start( s.preferACL.widget() )
        s.sourceHB.show_all()
        s.preferHB.show_all()
        
    def addOrder(s,b):
        newOrder = DNS.ACL({},s.attr.nmdc)
        n = s.orderList.iter_n_children(None)
        iter = s.orderList.append([ _('Sort Order')+ ' #' + str(n), newOrder ])
        s.attr.addrs.append(newOrder)
        s.orderTree.set_cursor(s.orderList.get_path(iter))
        s.editOrder(iter,newOrder)
        return False

    def removeOrder(s, b):
        if (s.currentIter != None):
            acl = s.orderList.get_value(s.currentIter,1)
            for i in range(0,len(s.attr.addrs)):
                if s.attr.addrs[i] == acl:
                    s.attr.addrs.pop(i)
            s.orderList.remove(s.currentIter)
            if s.orderList.iter_n_children(None):
                s.orderTree.set_cursor((0,))
                s.order_element_selected(s.orderTree,1)
            else:
                for c in s.sourceHB.get_children():
                    s.sourceHB.remove(c)
                for c in s.preferHB.get_children():
                    s.preferHB.remove(c)                    
                s.currentACL = None
                s.currentIter = None
                s.sourceHB.show_all()
                s.preferHB.show_all()
        return False
    
    def order_element_activated(s, tree, path, column):
        (iter,acl) = s.selected(s.orderTree)
        if acl != None:
            s.editOrder(iter,acl)
        return False
    
    def order_element_selected(s, tree, event):        
        (iter,acl) = s.selected(s.orderTree)
        if acl != None:
            s.editOrder(iter,acl)
        return False
    
    def load(s):
        i=0
        s.orderList.clear()
        for acl in s.attr.addrs:
            s.orderList.append([ _('Sort Order') + ' #' + str(i), acl ])
            i+=1
        if i>0:
            s.orderTree.set_cursor((0,))
            s.order_element_selected(s.orderTree,1)
    
    def value(s):
        return s.attr.out()

    def saveOrder(s, model, path, iter, user_data=None):
        s.attr.addrs.append(model.get_value(iter,1))
        
    def save(s):
        if s.sourceACL != None:
            s.sourceACL.save()
        if s.preferACL != None:
            s.preferACL.save()
        s.attr.addrs = []
        s.orderList.foreach(s.saveOrder)
        
    def widget(s):
        s.hb = gtk.HBox()
        s.orderList = gtk.ListStore( gobject.TYPE_STRING, gobject.TYPE_PYOBJECT )
        s.orderTree = gtk.TreeView( model=s.orderList )
        s.orderTree.set_headers_visible(False)
        s.orderTree.connect('row_activated',s.order_element_activated)
        s.orderTree.connect('button_release_event',s.order_element_selected)
        s.orderCell = gtk.CellRendererText()
        s.orderColm = gtk.TreeViewColumn( None, s.orderCell, text=0 )
        s.orderTree.append_column(s.orderColm)
        s.orderSW = gtk.ScrolledWindow()
        s.orderSW.set_policy(gtk.POLICY_AUTOMATIC,gtk.POLICY_AUTOMATIC)
        s.orderSW.add_with_viewport(s.orderTree)
        s.hb.pack_start(s.orderSW)
        s.sourceFrame = gtk.Frame(_('Source Address Match List'))
        s.preferFrame = gtk.Frame()
        s.preferLabel = gtk.Label(_('RR Set Order Address Match List'))
        s.preferFrame.set_label_widget(s.preferLabel)
        s.orderVB=gtk.VBox()
        s.sourceHB=gtk.HBox()
        s.preferHB=gtk.HBox()
        s.sourceFrame.add(s.sourceHB)
        s.preferFrame.add(s.preferHB)
        s.orderVB.pack_start(s.sourceFrame)
        s.orderVB.pack_start(s.preferFrame)
        s.butHB=gtk.HBox()
        s.addButton =   gtk.Button(stock=gtk.STOCK_ADD)
        s.addButton.connect('clicked',s.addOrder)
        vb = gtk.VBox()
        vb.pack_start(s.addButton,False)
        s.butHB.pack_start(vb,False)
        s.removeButton =   gtk.Button(stock=gtk.STOCK_REMOVE)
        s.removeButton.connect('clicked',s.removeOrder)
        vb = gtk.VBox()
        vb.pack_start(s.removeButton,False)
        s.butHB.pack_start(vb,False)
        s.orderVB.pack_start(s.butHB,False)
        s.hb.pack_start(s.orderVB)
        s.currentACL=None
        s.aclValues={ 'values':[_('IPv4 Address'), _('IPv6 Address'),
                                _('Local Host'),   _('Local Networks'),
                                _('Named ACL'),    _('ACL')
                               ]
                    }
        s.load()
        s.hb.show_all()
        return s.hb
    
class DNS_Path(Widget):
    def __init__(s, d, r, a, k):
        Widget.__init__(s, d, r, a, k)
        if( type(s.rr) == dict ) and s.rr.has_key('isDirectory'):
            s.isDirectory=true

    def get_filename(s):
        file=s.fileEntry.get_text()
        debug( 'G FILE:',s.filename)
        d = s.dialog.bind.nmdc.directory()
        abs = ((type(s.rr)==dict) and (s.rr.has_key('isAbsolute')))
        if (len(file) == 0):
            if( abs ):
                file='/'
            else:
                file=d
        if (not abs) and (file != '/'):
            file = d +'/'+file
        debug( 'G FILE:',s.filename)
        return file

    def destroy(s):
        pass
    
    def entry_focus_out(s, entry, event, problem=None):
        debug( 'NEW FILE:', s.fileEntry.get_text())
        s.fc_response(s,gtk.RESPONSE_ACCEPT,problem)
        return False

    def check_file(s,problem):
        s.entry_focus_out(s.fileEntry,1,problem)
        return not problem.has_key('yes')
    
    def fc_response(s, fc, response,problem=None):
        file=fc.get_filename()
        debug( 'FC FILE:',file)
        d = s.dialog.bind.nmdc.directory()
        cd = s.dialog.bind.nmdc.chroot
        abs = ((type(s.rr)==dict) and (s.rr.has_key('isAbsolute')))
        if (file == s.filename) or (response != gtk.RESPONSE_ACCEPT):
            fc.destroy()
            debug('FC SAME');
            return False
        else:            
            if (os.access(file, os.F_OK)==1):
                debug('FC EXISTS');
                st=os.stat(file)
                if (not stat.S_ISREG(st.st_mode)) and (not hasattr(s,'isDirectory')):
                    s.dialog.bind.gui.modalError(s.dialog.editDialog,None ,_('ERROR: Existing path is not regular file.'),fatal=False)
                    if( problem != None):
                        problem['yes']=True
                    s.warned = False
                    return False
                if hasattr(s,'isDirectory') and (not stat.S_ISDIR(st.st_mode)):
                    s.dialog.bind.gui.modalError(s.dialog.editDialog,None ,_('ERROR: Existing path is not directory.'),fatal=False)
                    if( problem != None):
                        problem['yes']=True
                    s.warned = False
                    return False
                if not s.warned and not hasattr(s,'isDirectory'):
                    s.dialog.bind.gui.modalError(s.dialog.editDialog,None ,_('WARNING: File Exists and will be overwritten on save.'),fatal=False,warning=True) 
                    s.warned = True
                    if( problem != None):
                        problem['yes']=True
                    return False                
            else:
                debug('FC NOT EXISTS',file);
                ed = file.rfind('/')
                if ( ed > -1 ):
                    wdir = file[:ed]
                else:
                    wdir = d
                debug('FC WDIR',wdir);                    
                if (os.access(wdir, os.F_OK | os.W_OK)!=1):
                    s.dialog.bind.gui.modalError(s.dialog.editDialog,None,_('ERROR: Path is not writable'),fatal=False)
                    if( fc != s ):
                        fc.set_filename(s.filename)
                        fc.set_current_name(s.data)
                    if( problem != None):
                        problem['yes']=True
                    s.warned = False
                    return False        
        if( fc != s ):
            fc.destroy()
        if( len(file) == 0 ):
            file = d
        if not abs:
            if (len(file) > len(d)) and (file[:len(d)] == d):
                file=file[len(d)+1:]
            elif(len(file) > len(cd)) and (file[:len(cd)] == cd):
                file=file[len(cd):]
            if( len(file) == 0 ):
                file = d
        debug('FC FILE 2',file)
        s.filename = file
        s.data=file
        s.fileEntry.set_text(file)
        s.warned = False
        return False
                            
    def browse_clicked(s,b):
        #title=None, parent=None, action=gtk.FILE_CHOOSER_ACTION_OPEN, buttons=None, backend=None
        name=_('Choose File')
        if hasattr(s.dialog,'title'):
            name += ' '+_('for')+' '+s.dialog.title
        elif hasattr(s.dialog,'name') and (s.dialog.name != None):
            name += ' '+_('for')+' '+s.dialog.name
        actType = gtk.FILE_CHOOSER_ACTION_SAVE
        if hasattr(s,'isDirectory'):
            actType = gtk.FILE_CHOOSER_ACTION_OPEN
        s.fc=gtk.FileChooserDialog(title=name, parent=s.dialog.editDialog,
                                   action=actType,
                                   buttons=(gtk.STOCK_OK, gtk.RESPONSE_ACCEPT,
                                            gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL
                                           )
                                  )
        debug( 'BC FILE:',s.filename)
        d = s.dialog.bind.nmdc.directory()
        abs = ((type(s.rr)==dict) and (s.rr.has_key('isAbsolute')))
        if (len(s.data) == 0):
            if( abs ):
                s.data='/'
            else:
                s.data=d
        if (not abs) and (s.data[0] != '/'):
            s.filename=d +'/'+s.data
        else:
            s.filename = s.data
        debug( 'BC FILE: 2',s.filename)
        s.fc.set_filename(s.filename)
        basename=s.filename
        if( not( abs or hasattr(s,'isDirectory') )):
            pi=s.filename.rfind('/')
            if( pi > -1):
                basename=s.filename[pi+1:]
        s.fc.set_current_name(basename)                
        s.fc.connect('response',s.fc_response)
        s.warned=False
        s.fc.show()

    def value(s):
        d = s.dialog.bind.nmdc.directory()
        cd = s.dialog.bind.nmdc.chroot
        file=s.fileEntry.get_text()
        if( len(file) == 0 ):
            file = d
        if (len(file) > len(d)) and (file[:len(d)] == d):
            file=file[len(d)+1:]
        elif(len(file) > len(cd)) and (file[:len(cd)] == cd):
            file=file[len(cd):]
        return file
    
    def save(s):
        s.attr.path=s.value()
        
    def widget(s):
        s.hb = gtk.HBox()
        s.fileEntry = gtk.Entry()
        if( s.data == None) or (len(s.data)==0):
            if (type(s.rr)==dict) and (s.rr.has_key('default')):
                s.data = s.rr.has_key('default')
            else:
                s.data='.'
            s.attr.path=s.data
        s.fileEntry.set_text(s.data)
        s.filename = s.data
        s.entered=False
        s.fileEntry.connect('focus_out_event', s.entry_focus_out)
        s.button = gtk.Button(label='Browse...',stock=gtk.STOCK_OPEN)
        s.warned=False
        s.button.connect('clicked',s.browse_clicked)
        vb=gtk.VBox()
        xhb=gtk.HBox()
        xhb.pack_start(s.fileEntry)
        vb.pack_start(xhb,False)
        s.hb.pack_start(vb,False)
        vb=gtk.VBox()
        xhb=gtk.HBox()
        xhb.pack_start(s.button,False)
        vb.pack_start(xhb,False)
        s.hb.pack_start(vb,False)
        s.hb.show_all()
        return s.hb
        

class DNS_Boolean(Widget):
    def __init__(s, d, r, a, k):
        Widget.__init__(s, d, r, a, k)

    def toggle(s,b):
        if( s.attr.data == 1 ):
            s.attr.data = 0
            s.tb.set_active(False)
            s.tb.set_label(s.name + ' : ' + s.false)
        else:
            s.attr.data = 1
            s.tb.set_active(True)
            s.tb.set_label(s.name + ' : ' + s.true)

    def value(s):
        if s.tb.get_active() :
            return s.true
        return s.false

    def save(s):
        if s.tb.get_active() :
            s.attr.data = 1
        else:
            s.attr.data = 0        
    
    def widget(s):
        s.tb = gtk.ToggleButton()
        o=None
        if str(s.rr.__class__) == 'Options.Option':
            o=s.rr.option
            debug( 'BOOLEAN OPTION:', o)
        if ( o != None ) and o.has_key('values'):
            s.false = o['values'][0]
            s.true  = o['values'][1]
            s.name  = o['name']
            if (s.attr.data == None) and o.has_key('default'):
                s.attr.data = o['default']            
        else:
            s.false = 'no'
            s.true  = 'yes'
            s.name  = s.key
            if (s.attr.data == None):
               s.attr.data = 0
        state=s.false
        if s.attr.data == 1:
            state = s.true
            s.tb.set_active(True)
        else:
            s.tb.set_active(False)
        s.tb.set_mode(True)
        s.tb.set_label(s.name + ' : ' + state)
        s.tb.connect('toggled',s.toggle)
        s.hb = gtk.HBox()
        s.hb.pack_start(s.tb,False)
        s.hb.show_all()
        return s.hb

class DNS_Radio(Widget):
    def __init__(s, d, r, a, k):
        Widget.__init__(s, d, r, a, k)
        s.data=s.attr.data
        
    def changed(s,c):
        s.data = s.choice.get_active()
        debug( 'RADIO:',s.data)

    def value(s):
        s.data= s.choice.get_active()
        return s.attr.values[s.data]

    def save(s):
        s.attr.data = s.attr.values[s.data]
        
    def widget(s):
        s.choice = gtk.combo_box_new_text()
        debug( 'VALUES:', s.attr.__class__, s.attr.values)
        i=0
        for val in s.attr.values :
            s.choice.append_text(val)
            if(  ((type(s.data) == str) and (s.data==val))
              or (s.data == i)
              ):
                s.choice.set_active(i)
            i += 1
        s.choice.connect('changed',s.changed)
        s.hb=gtk.HBox()
        s.hb.pack_start(s.choice,False)
        s.hb.show_all()
        return s.hb

class DNS_Size(Widget):
    def __init__(s, d, r, a, k):
        Widget.__init__(s, d, r, a, k)

    def value_changed(s,w,e=None):        
        v=s.entry.get_text()
        scale=s.scales[ s.scalew.get_active() ]
        if len(v) <= 0:
            s.dialog.bind.gui.modalError(s.parent,_('Size Value Error'),_('An non-empty size string must be entered.'))
            s.entry.set_text(s.size)
            return False
        if v not in ['default','unlimited']:
            intValue=None
            try:
                intValue = int(v)
            except ValueError:
                intValue = None
            if intValue == None:
                s.dialog.bind.gui.modalError(s.parent,_('Size Value Error'),
                                             _('Invalid integer value.')
                                            )
                s.entry.set_text(str(s.size))
                return False            
            s.size = intValue
            s.scale = scale
        else:
            s.size = v
            s.scale=''
            s.scalew.set_active(0)
        return False
    
    def value(s):
        return str(s.size) + s.scale

    def save(s):
        s.value_changed(s,s.entry)
        v=s.value()
        if v != str(s.attr):
            s.attr.size=s.size
            s.attr.scale=s.scale
    
    def widget(s):
        s.hb = gtk.HBox()
        if str(s.rr.__class__) == 'Options.Option':
            s.option = s.rr.option
        else:
            s.option = {}
        if hasattr(s.dialog,'editDialog'):
            s.parent = s.dialog.editDialog
        else:
            s.parent = s.dialog.gui.gui
        s.size=str(s.attr.size)
        s.scale=s.attr.scale
        debug( 'SIZE WIDGET:',s.attr, s.size, s.scale)
        s.combo = gtk.combo_box_entry_new_text()
        s.combo.append_text('default')
        s.combo.append_text('unlimited')
        s.entry = s.combo.child
        if (s.size == None) or (len(s.size) <= 0):
            s.size='default'
            s.entry.set_text('default')
        else:
            s.entry.set_text(str(s.size))
        s.entry.connect('focus_out_event',s.value_changed)            
        s.hb.pack_start(s.combo,False)
        s.scales=['','K','M','G']
        s.scalew = gtk.combo_box_new_text()
        s.scalew.append_text('')
        s.scalew.append_text('Kilo-Bytes')
        s.scalew.append_text('Mega-Bytes')
        s.scalew.append_text('Giga-Bytes')
        if (s.scale == None) or (len(s.scale)<=0):
            s.scale=''
            s.scalew.set_active(0)
        else:
            for k in range(0,len(s.scales)):
                if s.scale == s.scales[k]:
                    s.scalew.set_active(k)                
        s.scalew.connect('changed',s.value_changed)
        s.hb.pack_start(s.scalew,False,False)
        s.hb.show_all()
        return s.hb

class DNS_Key(Widget):
    def __init__(s, d, r, a, k):
        Widget.__init__(s, d, r, a, k)
        s.bind = d.bind
        s.gui = d.bind.gui

    def newSecret(s,b):
        r = s.bind.newKey()
        debug( 'newSecret', r)
        if type(r) == str:
            s.sLabel.set_markup('<tt>'+r+'</tt>')
        else:
            s.bind.gui.modalError(s.bind.gui.gui,_('DNS Security Key Generation Failed:'),_(r[1]))

    def widget(s):
        s.vb=gtk.VBox()
        if (s.attr.name == None) or (len(s.attr.name) == 0):
            s.nFrame = gtk.Frame(_('Key Name'))
            s.entry = gtk.Entry()
            s.nFrame.add(s.entry)
            s.vb.pack_start(s.nFrame)
        s.aFrame=gtk.Frame(_('Algorithm'))
        s.aLabel=gtk.Label(s.attr.algorithm) # hmac-md5 is ONLY alg supported by BIND9 - so far!
        s.aFrame.add(s.aLabel)
        s.sFrame=gtk.Frame(_('Secret'))
        s.sLabel=gtk.Label()
        s.sLabel.set_markup('<tt>'+s.attr.secret+'</tt>')
        s.sFrame.add(s.sLabel)
        s.vb.pack_start(s.aFrame)
        s.vb.pack_start(s.sFrame)
        s.newButton=gtk.Button(stock=gtk.STOCK_NEW,label=_('New Secret'))
        s.newButton.connect('clicked',s.newSecret)
        s.vb.pack_start(s.newButton,False)        
        return s.vb
    
    def value(s):
        return s.sLabel.get_text()

    def save(s):
        if (s.sLabel.get_text() != s.attr.secret):
            s.attr.secret = s.sLabel.get_text()
            return True
        return False
    
    def response(s,d,r):
        if(r == gtk.RESPONSE_OK):
            changed = s.save()            
            if (s.attr.name == None) or (len(s.attr.name)==0):
                name = s.entry.get_text()
                if len(name) <= 0:
                    s.bind.gui.modalError(s.bind.gui.gui,_('Remote Nameserver Control Security Key Name Error'),
                                          _('A non-empty Security Key Name is required.')
                                         )
                    return False
                if (s.bind.nmdc.__contains__(s.attr.key + [ "\"" + name + "\"" ]) or s.bind.nmdc.__contains__(s.attr.key + [ name ])):
                    s.bind.gui.modalError(s.bind.gui.gui,_('Remote Nameserver Control Security Key Name Error'),
                                          _('Key "%s" already exists.') % name
                                         )
                    return False
                if s.bind.acls.has_key(name) or (name.lower() in ['none','any','localhost','localnets']) :
                    s.bind.gui.modalError(s.bind.gui.gui,_('Remote Nameserver Control Security Key Name Error'),
                                          _('Name "%s" is an Access Control List name.') % name
                                         )
                    return False
                if(len(s.attr.secret) <= 0):
                    return False
                s.attr.name=name
                if not s.bind.nmdc.__contains__(s.attr.key):
                    s.bind.nmdc[ s.attr.key ] = s.attr.out()
                else:
                    s.bind.nmdc[ s.attr.key + [s.attr.name]] = s.attr.out()
                s.attr.key += [s.attr.name]
                s.bind.nmdc.update()
                s.bind.load()
            elif changed:                
                s.bind.nmdc[ s.attr.key ] = {'algorithm': s.attr.algorithm, 'secret': s.attr.secret}
                s.bind.nmdc.update()
        d.destroy()
        
    def keyDialog(s):
        t=''
        if (s.attr.name == None) or (len(s.attr.name)==0):
            t = _('New Remote Nameserver Control Security Key')
        else:
            t = RDNC_KEY_NAME+' '+s.attr.name
        s.editDialog = gtk.Dialog(parent=s.gui.gui, title=t,
                                  buttons=
                                  (gtk.STOCK_OK,     gtk.RESPONSE_OK,
                                   gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL
                                  )
                                 )
        s.editDialog.connect('response',s.response)
        s.editDialog.vbox.pack_start(s.widget())
        s.editDialog.show_all()
    
class int_widget(Widget):
    def __init__(s, d, r, a, k):
        Widget.__init__(s, d, r, a, k)

    def value(s):
        return str(int(s.adj.get_value()))

    def save(s):
        v = int(s.value())
        if type(s.rr) == dict:
            obj = s.rr.obj
            debug( 'DICT WIDGET INT ATTR SAVE',obj,s.key)
        elif (v !=  s.rr[s.key]):
            debug( 'INT',s.key,'SAVE',v    )
            setattr(s.rr,s.key,v)
            return True
        return False

    def widget(s):
        debug( 'INT WIDGET:',s.rr.__class__,s.key,s.attr)
        o = None            
        if str(s.rr.__class__) == 'Options.Option':
            o = s.rr.option
        elif type(s.rr) == dict:
            o = s.rr
            debug( 'INT dict OPTION:',o)
        else:
            debug( s.rr.__class__, s.rr)
        if( o != None):
            debug( 'INT OPTION:', o)
        if (s.attr == None):            
            if( o != None) and  o.has_key('default'):
                s.attr = o['default']
            else:
                s.attr = 0
        debug( 'INT WIDGET STILL:',s.attr)
        if( o != None ) and (o.has_key('range')):
            s.adj = gtk.Adjustment(value=float(s.attr), lower=float(o['range'][0]), upper=float(o['range'][1]), step_incr=1, page_incr=10,page_size=10)
            debug( 'RANGE:', o['range'][0], o['range'][1])
        else:
            s.adj = gtk.Adjustment(value=float(int(s.attr)), lower=0, upper=(2**32)-1, step_incr=1, page_incr=10,page_size=10)
        s.sb = gtk.SpinButton(adjustment=s.adj)
#        s.adj.set_value(float(int(s.attr)))
        s.sb.set_numeric(True)
        s.sb.set_value(float(int(s.attr)))
        s.hb = gtk.HBox()
        s.hb.pack_start(s.sb,False)
        debug( 'INT WIDGET FINAL:',s.attr)
        debug( 'INT WIDGET VALUE:',s.sb.get_value())
        s.hb.show_all()
        return s.hb

class dict_widget(Widget):
    def __init__(s, d, r, a, k):
        Widget.__init__(s, d, r, a, k)
        debug( 'DICT WIDGET NEW:',d,r,a,k)
        
    def value(s):
        v =''
        for attrd in s.rr.option['attrs']:
            for name, attr in attrd.items():
                if not s.widgets.has_key(name):
                    continue
                else:
                    w=s.widgets[name]
                v += w.value() +' '
        return v[:-1]

    def save(s):
        modified=False
        for attrd in s.rr.option['attrs']:
            for name, attr in attrd.items():
                if not s.widgets.has_key(name):
                    continue
                else:
                    w=s.widgets[name]
                v=w.value()
                debug( 'DICT WIDGET SAVE:',w.key,v)
                if( type(s.attr[w.key]) in [ int, str ] ):
                    if( s.attr[w.key] != v ):
                        s.attr[w.key] = w.value()
                        modified=True
                else:
                    if str(s.rr.value[w.key]) != v:
                        w.save()
                        modified=True
        return modified

    def add_attr(s,b,a):
        name,attr,container=a
        debug( 'ADD ATTR:',name,attr,container)
        container.hide()
        for c in container.get_children():
            container.remove(c)
        if attr.has_key('default'):
            v=attr['default']
        else:
            v=None
        if( attr['type'] == DNS.ACL):
            s.attr[name]=attr['type'](v, s.dialog.bind.nmdc, s.key)
        elif(attr['type'] == DNS.Radio):
            s.attr[name]=attr['type'](v,attr['values'])
        elif(attr['type'] == DNS.IPV4.Address):
            s.attr[name]=DNS.IPV4.Address(v,input=True,prefixAllowed=(v.find('/')>0))
        elif(attr['type'] == DNS.IPV6.Address):
            s.attr[name]=DNS.IPV6.Address(v,allow_prefix=(v.find('/')>0))
        else:
            s.attr[name]=attr['type'](v)
        w = s.gui.widget(s.attr[name])(s.dialog, attr, s.attr[name], name )
        s.widgets[name]=w
        w.attrFrame=container
        vb=gtk.VBox()
        vb.pack_start(w.widget(),False)
        if not attr.has_key('required'):
            hb=gtk.HBox()
            removeButton = gtk.Button(stock=gtk.STOCK_REMOVE)
            removeButton.connect('clicked',s.remove_attr,[name,attr,container])
            hb.pack_start(removeButton,False)
            vb.pack_start(hb,False)
        container.add(vb)
        container.set_label(name)
        container.show_all()
        
    def remove_attr(s,b,a):
        name,attr,container=a
        debug( 'REMOVE ATTR:',attr)
        container.hide()
        for c in container.get_children():
            container.remove(c)
        del s.widgets[name]
        del s.attr[name]
        hb=gtk.HBox()
        addButton=gtk.Button(stock=gtk.STOCK_ADD)
        addButton.connect('clicked',s.add_attr,[name,attr,container])
        hb.pack_start(addButton,False)
        container.add(hb)
        container.set_label(name)
        container.show_all()
        
    def widget(s):
        s.vb=gtk.VBox()
        gui=s.dialog.bind.gui
        s.gui=gui
        s.widgets={}
        n=0
        for attrd in s.rr.option['attrs']:
            for name, attr in attrd.items():
                debug( 'DICT widget attr:', name, s.attr[name].__class__)# s.attr[attr])
                if(  s.attr[name] == None ):
                    debug( 'OMITTED ATTR:',name)
                    fm=gtk.Frame(label=name)
                    hb=gtk.HBox()
                    addButton=gtk.Button(stock=gtk.STOCK_ADD)
                    addButton.connect('clicked',s.add_attr,[name,attr,fm])
                    hb.pack_start(addButton,False)
                    fm.add(hb)
                    s.vb.pack_start(fm)
                    continue                
                w = gui.widget(s.attr[name])(s.dialog, attr, s.attr[name], name )
                s.widgets[name]=w
                w.attrFrame = gtk.Frame(label=name)
                vb=gtk.VBox()            
                vb.pack_start(w.widget(),False)
                if not attr.has_key('required'):
                    hb=gtk.HBox()
                    removeButton = gtk.Button(stock=gtk.STOCK_REMOVE)
                    removeButton.connect('clicked',s.remove_attr,[name,attr,w.attrFrame])
                    hb.pack_start(removeButton,False)
                    vb.pack_start(hb,False)
                w.attrFrame.add(vb)
                s.vb.pack_start(w.attrFrame)
        s.vb.show_all()
        return s.vb                


class DNS_AddressList(Widget):
    def __init__(s, d, r, a, k):
        Widget.__init__(s, d, r, a, k)
        s.bind = d.bind

    def selected(s,tree):
        selection = tree.get_selection()
        if( selection != None ):
            model, selected = selection.get_selected()
            if (model != None) and  (selected != None):
                return (selected,) + model.get(selected, 0, 1)
        return (None,None,None)

    def remove_addr(s,b,iter):
        s.addrList.remove(iter)
        if s.addrList.iter_n_children(None) > 0:
            s.addrTree.set_cursor((0,))
        s.addr_element_change(s.addrTree,1)
            
    def save_addr(s,b,iter):
        a=s.addrWidget.value()
        addr=DNS.Address(a)
        if s.portWidget != None:
            addr.port = s.portWidget.value()
        s.addrList.set( iter, 0, a, 1, addr )

    def add_addr(s,b):
        a = s.addrWidget.value()
        addr=DNS.Address(a)
        if s.portWidget != None:
            addr.port = s.portWidget.value()
        iter = s.addrList.append([ a, addr ])
        s.addrTree.set_cursor(s.addrList.get_path(iter))
        s.addr_element_change( s.addrTree,1)

    def add_port(s,b):
        s.portContainer.hide()
        s.portContainer.remove(s.portButton)
        s.port=53
        s.portWidget = int_widget(s,s,53,'port')
        s.portContainer.pack_start(s.portWidget.widget())
        s.portButton=gtk.Button(stock=gtk.STOCK_REMOVE,label=_('Remove Port'))
        s.portButton.connect('clicked',s.remove_port)
        s.portContainer.pack_start(s.portButton)
        s.portContainer.show_all()
        
    def remove_port(s,b):
        s.portContainer.hide()
        for c in s.portContainer.get_children():
            s.portContainer.remove(c)
        s.portWidget = None
        s.portButton=gtk.Button(stock=gtk.STOCK_ADD,label=_('Add Port'))
        s.portButton.connect('clicked',s.add_port)
        s.portContainer.pack_start(s.portButton)
        s.portContainer.show_all()
                                   
    def edit_addr_element(s, attr, iter):
        editContainer = gtk.VBox()
        s.eHB=gtk.HBox()        
        s.okButton = gtk.Button(stock=gtk.STOCK_OK)
        debug( 'EDIT:',attr.__class__,iter)
        if iter != None:
            debug( 'EDIT ADDR',attr.__class__,iter)
            s.addrWidget = s.bind.gui.widget(attr.address)(s, s, attr.address, None)
            s.removeButton = gtk.Button(stock=gtk.STOCK_REMOVE)
            s.removeButton.connect('clicked',s.remove_addr,iter)
            s.eHB.pack_start(s.removeButton)
            s.okButton.connect('clicked',s.save_addr,iter)
        else:
            debug( 'EDIT NEW ADDR:',attr.__class__)
            new_attr=copy.deepcopy(attr)
            s.addrWidget = s.bind.gui.widget(new_attr)(s, s, new_attr, None)
            s.okButton.connect('clicked',s.add_addr)
        s.eHB.pack_start(s.okButton)
        s.addrFrame = gtk.Frame(label=_('Address'))
        s.addrFrame.add(s.addrWidget.widget())
        editContainer.pack_start(s.addrFrame,False)
        s.portFrame=gtk.Frame(label=_('Port'))
        s.portContainer=gtk.VBox()
        if hasattr( attr, 'port') and (attr.port != None):
            s.portWidget = int_widget(s,attr,attr.port,'port')            
            s.portContainer.pack_start(s.portWidget.widget(),False)
            s.portButton=gtk.Button(stock=gtk.STOCK_REMOVE,label=_('Remove Port'))
            s.portButton.connect('clicked',s.remove_port)
            s.portContainer.pack_start(s.portButton,False)
        else:
            s.portWidget=None
            s.portButton=gtk.Button(stock=gtk.STOCK_ADD,label=_('Add Port'))
            s.portButton.connect('clicked',s.add_port)
            s.portContainer.pack_start(s.portButton)
        s.portFrame.add(s.portContainer)
        editContainer.pack_start(s.portFrame)
        editContainer.pack_start(s.eHB)
        return editContainer
    
    def addr_element_change(s, tree, new=None):
        iter,ads,a=s.selected(tree)
        if s.editContainer != None:
            s.editContainer.hide()
            s.editHB.remove(s.editContainer)
            s.editContainer.destroy()
        if( new == None):
            iter = None
        if( a != None):
            attr=a
            s.editContainer = s.edit_addr_element(attr,iter)
            s.editHB.pack_start(s.editContainer)
        else:
            s.editContainer = None
        s.editHB.show_all()

    def addr_element_activated(s,tree,path,column):
        s.addr_element_change(tree,path)
        return False
    
    def new_addr_element_activated(s,tree,path,column):
        s.addr_element_change(tree)
        return False
    
    def addr_element_selected(s, tree, event):
        s.addr_element_change(tree,1)
        return False

    def new_addr_element_selected(s, tree, event):
        s.addr_element_change(tree)
        return False

    def load(s):
        s.addrList.clear()
        debug( 'AddressList: LOAD')
        for a in s.attr.addrs:
            debug( '   ADDR:', str(a), hasattr(a,'port'))
            s.addrList.append([str(a.address),a])
    
    def addr_elements_value(s, model, path, iter, user_data=None):
        ele=model.get_value(iter, 1)
        v=str(ele)
        if(hasattr(ele,'port')):
            v += 'port ' + str(ele.port)
        v+='; '
        s.addr_elements_string += v
        return False                            
    
    def value(s):
        s.addr_elements_string = ''
        s.addrList.foreach(s.addr_elements_value)
        debug( 'AddressList VALUE:',s.addr_elements_string)
        return s.addr_elements_string

    def save_addr_element(s, model, path, iter, user_data=None):
        ele = model.get_value(iter, 1)
        s.attr.addrs.append( ele )
        debug( 'AddressList ELEMENT SAVED:', str(ele))
        return False                            
        
    def save(s):
        debug( 'AddressList SAVE ',s.rr.__class__)
        s.attr.addrs=[]
        s.addrList.foreach(s.save_addr_element)
    
    def widget(s):
        debug( 'AddressList Widget')
        s.hb=gtk.HBox()
        s.addrList = gtk.ListStore( gobject.TYPE_STRING, gobject.TYPE_PYOBJECT )
        s.newList = gtk.ListStore( gobject.TYPE_STRING, gobject.TYPE_PYOBJECT )
        s.addrTree = gtk.TreeView( model=s.addrList )
        s.newTree = gtk.TreeView( model=s.newList )
        s.addrTree.set_headers_visible(False)
        s.newTree.set_headers_visible(False)
        s.addrSW =   gtk.ScrolledWindow()
        s.addrSW.set_policy(gtk.POLICY_AUTOMATIC,gtk.POLICY_AUTOMATIC)
        s.newSW =   gtk.ScrolledWindow()
        s.newSW.set_policy(gtk.POLICY_AUTOMATIC,gtk.POLICY_AUTOMATIC)
        s.addrTree.connect('row_activated',s.addr_element_activated)
        s.newTree.connect('row_activated',s.new_addr_element_activated)
        s.addrTree.connect('button_release_event',s.addr_element_selected)
        s.newTree.connect('button_release_event',s.new_addr_element_selected) 
        s.addrCell = gtk.CellRendererText()
        s.newCell = gtk.CellRendererText()
        s.addrColm = gtk.TreeViewColumn( None, s.addrCell, text=0 )
        s.newColm = gtk.TreeViewColumn( None, s.newCell, text=0 )
        s.addrTree.append_column(s.addrColm)
        s.newTree.append_column(s.newColm)        
        s.addrListFrame = gtk.Frame(label=_('  Address List '))
        s.addrSW.add_with_viewport(s.addrTree)
        s.addrListFrame.add(s.addrSW)
        s.editFrame = gtk.Frame(label=_('Edit List Element'))        
        s.editHB = gtk.HBox()
        s.editContainer= gtk.VBox()
        s.addrWidget = gtk.Label('')
        s.portWidget = None
        s.editContainer.pack_start(s.addrWidget)
        s.editHB.pack_start(s.editContainer)
        s.editFrame.add(s.editHB)
        s.newFrame = gtk.Frame(label=_('New List Element'))
        s.newSW.add_with_viewport(s.newTree)
        s.newFrame.add(s.newSW)
        s.hb.pack_start(s.addrListFrame)
        s.hb.pack_start(s.editFrame)
        s.hb.pack_start(s.newFrame)
        s.newList.append([_('IPV4 Address'), DNS.IPV4.Address('0.0.0.0',True,False)])
        s.newList.append([_('IPV6 Address'), DNS.IPV6.Address('::',DNS.IPV6.HEX_FORMAT,False)])
        s.load()
        s.hb.show_all()
        return s.hb

class DNS_Controls(Widget):
    def __init__(s, d, r, a, k):
        debug( 'DNS_Controls INIT')
        s.dialog=d
        s.rr=r
        s.attr=a
        s.key=k
        s.gui=d.gui
        s.bind=d.bind
        s.parent = s.bind.gui.gui
        
    def save_inet(s):
        debug( 'SAVE INET')
        if hasattr(s,'inet') and ( s.inet != None):
            s.value=s.inet
            s.listenW.save()
            if( s.inet['port'] != None ):
                debug( 'SAVE PORT:',s.inet['port'])
                s.inet['addr'].port=int(s.inet['port'])
                s.inet['port']=int(s.inet['port'])
            s.addrW.save()
            s.keysW.save()

    def add_inet(s, b):
        debug( 'ADD INET')
        new_inet={ 'addr': DNS.IPV4.Address('0.0.0.0'),
                   'allow': DNS.ACL(None,s.bind.nmdc,None),
                   'keys':  DNS.ACL(None,s.bind.nmdc,name=None,isControlKeys=True)
                 }
        iter = s.addrList.append(['0.0.0.0', new_inet])
        s.addrTree.set_cursor(s.addrList.get_path(iter))
        s.select_control(new_inet,iter)
        
    def remove_inet(s,b,iter=None):
        if (iter != None) and (s.addrList.iter_n_children(None) > 1):
            debug( 'REMOVING ITER:',iter)
            s.addrList.remove(iter)
            iter = s.addrList.get_iter((0,))
            s.select_control(s.addrList.get_value(iter,1),iter)
        else:
            s.bind.gui.modalError(s.parent,_('Controls Element Remove Error'),_('Element cannot be removed: the controls object must contain at least one element.'))
                                  
    def select_control(s,inet,iter):
        debug( 'SELECT_CONTROL INET:',inet)
        s.save_inet()
        s.inet = inet        
        if (s.inet != None) :
            if not(hasattr(s.inet['addr'],'port')):
                s.inet['port']=None
            else:
                s.inet['port']=int(s.inet['addr'].port)
        s.listenHB.remove(s.listenWidget)
        s.listenW = dict_widget(s,s,inet,'addr')
        s.listenWidget = s.listenW.widget()
        s.listenHB.pack_start(s.listenWidget)
        s.addrHB.remove(s.addrWidget)
        s.addrW=DNS_ACL(s,
                        { 'values': [   'IPv4 Address', 'IPv6 Address',
                                        'Local Host', 'Local Networks',
                                        'Named ACL','ACL'
                                    ]
                        },
                        inet['allow'],
                        'allow'
                       )
        s.addrWidget = s.addrW.widget()
        s.addrHB.pack_start(s.addrWidget)
        s.keysHB.remove(s.keysWidget)
        s.keysW=DNS_ACL(s,
                        {'values': [ 'Named Key' ] },
                        inet['keys'],
                        'keys'
                       )
        s.keysWidget=s.keysW.widget()
        s.keysHB.pack_start(s.keysWidget)
        s.removeButton.disconnect(s.removeHandler)
        debug( 'ITER FOR REMOVE: ',iter)
        s.removeHandler=s.removeButton.connect('clicked',s.remove_inet,iter)
            
    def selected(s,tree):
        selection = tree.get_selection()
        if( selection != None ):
            model, selected = selection.get_selected()
            if (model != None) and  (selected != None):
                return (selected,) + model.get(selected, 1)
        return (None,None)
    
    def addr_element_activated(s,t,p,c):
        (iter,inet)=s.selected(s.addrTree)
        debug( 'SELECTED INET:',inet)
        s.select_control(inet,iter)
        
    def addr_element_selected(s,t,e):
        (iter,inet)=s.selected(s.addrTree)
        debug( 'SELECTED INET:',inet)
        s.select_control(inet,iter)
        
    def load(s):
        s.addrList.clear()
        firstNode=None
        firstInet=None
        path=0
        for inet in s.attr.inets:
            ic=copy.deepcopy(inet)
            debug( 'INET:',ic)
            node = s.addrList.append([str(inet['addr']),ic])
            if firstNode == None:
                firstNode = node
                firstInet = ic
            if hasattr(ic['addr'],'port'):
                ic['port']= int(inet['addr'].port)
            path += 1
        if firstNode != None:
            s.addrTree.set_cursor((path,))
            s.inet = None
            s.select_control(firstInet,firstNode)
            
    def widget(s):
        debug( 'DNS_Controls Widget')
        s.option ={ 'attrs':
                    [ { 'addr':
                       { 'name': _('Address'),
                         'type': DNS.IPV4.Address,
                         'syntax': True,
                         'default': socket.gethostbyname(socket.getfqdn()),
                         'required': True
                       }
                     },
                     { 'port':
                       { 'name': _('Port'),
                         'type': int,
                         'range': [0,65535],
                         'syntax': True,
                         'default': 953
                       }
                     }
                    ]
                  }
        s.gui=s.dialog.gui
        s.bind=s.dialog.bind
        s.hb=gtk.HBox()
        s.addrList = gtk.ListStore( gobject.TYPE_STRING, gobject.TYPE_PYOBJECT )
        s.addrTree = gtk.TreeView( model=s.addrList )
        s.addrTree.set_headers_visible(False)
        s.addrTree.connect('row_activated',s.addr_element_activated)
        s.addrTree.connect('button_release_event',s.addr_element_selected)
        s.addrCell = gtk.CellRendererText()
        s.addrColm = gtk.TreeViewColumn( None, s.addrCell, text=0 )
        s.addrTree.append_column(s.addrColm)
        s.addrSW = gtk.ScrolledWindow()
        s.addrSW.set_policy(gtk.POLICY_AUTOMATIC,gtk.POLICY_AUTOMATIC)
        s.addrSW.add_with_viewport(s.addrTree)        
        s.hb.pack_start(s.addrSW)
        s.vb=gtk.VBox()
        s.listenFrame= gtk.Frame(_('Address to Listen On'))
        s.listenHB=gtk.HBox()
        s.listenWidget=gtk.Label("")
        s.listenHB.pack_start(s.listenWidget)
        s.listenFrame.add(s.listenHB)
        s.vb.pack_start(s.listenFrame)
        s.addrFrame= gtk.Frame(_('Addresses to Allow'))
        s.addrHB=gtk.HBox()
        s.addrWidget=gtk.Label("")
        s.addrHB.pack_start(s.addrWidget)
        s.addrFrame.add(s.addrHB)
        s.vb.pack_start(s.addrFrame)
        s.keysFrame = gtk.Frame(_('Security Keys to Allow'))
        s.keysHB=gtk.HBox()
        s.keysWidget=gtk.Label("")
        s.keysHB.pack_start(s.keysWidget)
        s.keysFrame.add(s.keysHB)
        s.vb.pack_start(s.keysFrame)
        s.butHB=gtk.HBox()
        s.addButton=gtk.Button(stock=gtk.STOCK_ADD)
        s.addButton.connect('clicked',s.add_inet)
        s.butHB.pack_start(s.addButton)
        s.removeButton=gtk.Button(stock=gtk.STOCK_REMOVE)
        s.removeHandler=s.removeButton.connect('clicked',s.remove_inet)
        s.butHB.pack_start(s.removeButton)
        s.vb.pack_start(s.butHB)
        s.hb.pack_start(s.vb)
        s.load()
        return s.hb

    def save_ele(s, model, path, iter, user_data=None):
        inet = model.get_value(iter, 1)
        s.attr.inets.append( inet )
        return False

    def save(s):
        s.save_inet()
        s.attr.inets=[]
        s.addrList.foreach(s.save_ele)
        if s.bind.nmdc.__contains__('controls'):
            del s.bind.nmdc['controls']            
        s.bind.nmdc['controls']=s.attr.out()
        s.bind.nmdc.update()
        s.bind.zoneTree.load()
            
    def response(s,d,r):
        if r == gtk.RESPONSE_OK:
            s.save()
        d.destroy()
        
    def ctlDialog(s):
        debug( 'DNS_Controls DIALOG')
        s.ctlDialogWidget = gtk.Dialog(parent=s.gui.gui, title=_('DNS Controls'),
                                  buttons=
                                  (gtk.STOCK_OK,     gtk.RESPONSE_OK,
                                   gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL
                                  )
                                 )
        s.parent=s.ctlDialogWidget
        s.ctlDialogWidget.connect('response',s.response)
        s.ctlDialogWidget.vbox.pack_start(s.widget())
        s.ctlDialogWidget.show_all()
#
# Logging:
#
class DNS_Logging(Widget):
    def __init__(s, d, r, a, k):
        debug( 'DNS_Logging INIT')
        s.dialog=d
        s.rr=r
        s.attr=copy.deepcopy(a)
        s.key=k
        s.gui=d.gui
        s.bind=d.bind
        s.parent = s.bind.gui.gui
        s.editDialog = s.parent

    def saveChan(s):
        debug( 'LOGGING: saveChan:',s.channel.name)
        if s.channel.name in ['default_debug','default_syslog','default_stderr','null']:
            debug( '  NOT saving',s.channel.name)
            return True
        if (s.channel.name != s.chanName.get_text()):
            p={}
            s.checkChanName(s.chanName, 1, p)
            if p.has_key('problem'):
                return False
            old_name = s.channel.name
            s.channel.name = s.chanName.get_text()
            if s.attr.channels.has_key(old_name):
                del s.attr.channels[old_name]
            s.attr.channels[s.channel.name] = s.channel
            s.chanList.set_value(s.chanIter,0,s.channel.name)
            debug( '  SAVED channel', s.channel.name)
        if (s.channel.dest == 'file'):
            p={}
            if (not hasattr(s.fileWidget,'warned')) or (s.fileWidget.warned == False):                
                if not s.fileWidget.check_file(p):
                    s.fileWidget.fileEntry.grab_focus()
                    return False
            if hasattr(s.channel,'file'):
                debug( 'CHAN FILE:', s.channel.file.path.__class__, s.channel.file.path.path)
            if not(hasattr(s.channel,'file')):
                s.channel.file = DNS.Logging.File( s.fileWidget.value() )
            elif(s.channel.file.path.path != s.fileWidget.value()):
                s.channel.file.path.path = s.fileWidget.value()
            if hasattr(s,'versionSB'):
                if s.verUnlimited.get_active():
                    s.channel.file.versions = 'unlimited'
                else:
                    s.channel.file.versions = int( s.versionSB.get_value() )
            elif hasattr(s.channel.file,'versions'):
                del s.channel.file.versions
            if hasattr(s,'sizeWidget'):
                if (not hasattr(s.channel.file,'size')) or (str(s.channel.file.size) != s.sizeWidget.value()):
                    s.channel.file.size = DNS.Size(s.sizeWidget.value())
            elif hasattr(s.channel.file,'size'):
                del s.channel.file.size
        elif (s.channel.dest == 'syslog'):
            s.channel.syslog = DNS.Logging.syslog_facilities[s.logFacilityChoice.get_active()]
        if hasattr(s,'severityChanged'):
            s.channel.severity = DNS.Logging.severities[ s.severityChoice.get_active() ]
        if s.channel.opts.has_key('print-time'):
            if not s.optPrintTime.get_active():
                del s.channel.opts['print-time']
            else:
                s.channel.opts['print-time'] = 'yes'
        elif s.optPrintTime.get_active():
            s.channel.opts['print-time'] = 'yes'        
        if s.channel.opts.has_key('print-severity'):
            if not s.optPrintSeverity.get_active():
                del s.channel.opts['print-severity']
            else:
                s.channel.opts['print-severity'] = 'yes'
        elif s.optPrintSeverity.get_active():
            s.channel.opts['print-severity'] = 'yes'
        if s.channel.opts.has_key('print-category'):
            if not s.optPrintCategory.get_active():
                del s.channel.opts['print-category']
            else:
                s.channel.opts['print-category'] = 'yes'
        elif s.optPrintCategory.get_active():
            s.channel.opts['print-category'] = 'yes'            
        return True
        
    def checkChanName(s,w,e,p=None):
        n = s.chanName.get_text()
        if p == None:
            p = {}
        if (s.channel.name != n) and s.attr.channels.has_key(n) and (s.channel != s.attr.channels[n]):
            p['problem']=_('Channel "%s" already exists.') % n
        elif (len(n) == 0):
            p['problem']=_('The channel name cannot be empty.')
        if p.has_key('problem'):
            s.dialog.gui.modalError(s.parent,_('Channel Name Error'), p['problem'])
            s.chanName.set_text(s.channel.name)
            s.chanTree.set_cursor(s.chanList.get_path(s.chanIter))
        else:
            s.chanList.set_value(s.chanIter,0,n)
        return False
        
    def addChan(s,b):
        chan = DNS.Logging.Channel(None,None)
        n = s.chanList.append(['',chan])
        s.chanTree.set_cursor(s.chanList.get_path(n))
        chan.new_channel=1
        s.editChan(n,chan)
    
    def removeChan(s,b):
        s.chanList.remove(s.chanIter)
        if s.attr.channels.has_key(s.channel.name):
            del s.attr.channels[ s.channel.name ]
        s.channel = None
        s.chanIter= None
        if s.chanList.iter_n_children(None) > 0:
            s.chanTree.set_cursor((0,))
            s.chan_element_selected(s.chanTree,1)
        else:
            for c in  s.chanVB.get_children():
                s.chanVB.remove(c)
    
    def changeDest(s,w):
        newDest =  s.attr.destinations[ s.destChoice.get_active() ]
        if newDest == s.channel.dest:
            return
        debug( 'CHANGE DEST:', s.channel.dest, newDest)
        if s.channel.dest == 'file':
            s.destHB.remove(s.fileFrame)
            if hasattr(s.channel,'file'):
                del s.channel.file
        elif s.channel.dest == 'syslog':
            s.destHB.remove(s.logFacilityChoice)
            if hasattr(s.channel,'syslog'):
                del s.channel.syslog
        s.channel.dest = newDest
        if newDest == 'file':
            setattr(s.channel,newDest,DNS.Logging.File(''))
        elif newDest == 'syslog':
            setattr(s.channel,newDest,'daemon')
        s.set_dest()
        s.destHB.show_all()
        
    def changeSeverity(s,w):
        s.severityChanged = 1
    
    def addFileVersions(s,w,file=None):
        for c in s.verVB.get_children():
            s.verVB.remove(c)
        s.channel.file.versions='unlimited'
        s.setVersionFrame(s.channel)        
        s.verVB.show_all()

    def removeFileVersions(s,w,file=None):
        for c in s.verVB.get_children():
            s.verVB.remove(c)
        del s.channel.file.versions
        del s.versionSB
        s.setVersionFrame(s.channel)        
        s.verVB.show_all()

    def versionsUnlimited(s,w):
        if s.verUnlimited.get_active():
            s.versionSB.set_value(0.0)
            s.versionSB.set_sensitive(False)
        else:
            s.versionSB.set_sensitive(True)            
        
    def addFileSize(s,w,file=None):
        for c in s.sizeVB.get_children():
            s.sizeVB.remove(c)
        s.channel.file.size=DNS.Size('10M')
        s.setSizeFrame(s.channel)
        s.sizeVB.show_all()        

    def removeFileSize(s,w,file=None):
        for c in s.sizeVB.get_children():
            s.sizeVB.remove(c)
        del s.channel.file.size
        del s.sizeWidget
        s.setSizeFrame(s.channel)
        s.sizeVB.show_all()
    
    def setVersionFrame(s,chan):
        if( hasattr(chan.file,'versions')):
            v=chan.file.versions
            if chan.file.versions == 'unlimited':
                v=0
            vhb=gtk.HBox()
            vvb=gtk.VBox()
            s.versionAdj=gtk.Adjustment(value=float(int(v)), lower=0, upper=(2**32)-1, step_incr=1, page_incr=10,page_size=1)
            s.versionSB=gtk.SpinButton(adjustment=s.versionAdj)
            s.versionSB.set_numeric(True)                
            vvb.pack_start(s.versionSB,False)
            s.verUnlimited=gtk.CheckButton(label=_('Unlimited'))
            s.verUnlimited.connect('toggled',s.versionsUnlimited)
            if chan.file.versions == 'unlimited':
                s.versionSB.set_sensitive(False)
                s.verUnlimited.set_active(True)
            else:
                s.verUnlimited.set_active(False)                
            vvb.pack_start(s.verUnlimited,False)
            vhb.pack_start(vvb,False)
            s.removeVersionsButton = gtk.Button(stock=gtk.STOCK_REMOVE)
            s.removeVersionsButton.connect('clicked',s.removeFileVersions,chan.file)
            vvvb=gtk.VBox()
            vvvb.pack_start(s.removeVersionsButton,False)
            vhb.pack_end(vvvb,False)
            s.verVB.pack_start(vhb,False)
        else:
            s.addVersions=gtk.Button(stock=gtk.STOCK_ADD)
            s.addVersions.connect('clicked',s.addFileVersions,chan.file)            
            avhb=gtk.HBox()
            avhb.pack_start(s.addVersions,False)
            s.verVB.pack_start(avhb,False)
        
    def versionFrame(s,chan):
        verFrame=gtk.Frame(_('Versions'))
        s.verVB=gtk.VBox()
        s.setVersionFrame(chan)
        verFrame.add(s.verVB)
        return verFrame

    def setSizeFrame(s,chan):
        if hasattr(chan.file,'size'):
            v=chan.file.size
            rshb=gtk.HBox()
            s.sizeWidget = DNS_Size(s,chan.file,chan.file.size,'size')
            rshb.pack_start(s.sizeWidget.widget(),False)
            s.removeFileSizeButton = gtk.Button(stock=gtk.STOCK_REMOVE)
            s.removeFileSizeButton.connect('clicked',s.removeFileSize)
            rsvb=gtk.VBox()
            rsvb.pack_start(s.removeFileSizeButton,False)
            rshb.pack_end(rsvb,False)
            s.sizeVB.pack_start(rshb,False)
        else:
            s.addFileSizeButton=gtk.Button(stock=gtk.STOCK_ADD)
            s.addFileSizeButton.connect('clicked',s.addFileSize,chan.file)
            ashb=gtk.HBox()
            ashb.pack_start(s.addFileSizeButton,False)
            s.sizeVB.pack_start(ashb,False)
    
    def sizeFrame(s,chan):
        sizeFrame=gtk.Frame(_('Size'))
        s.sizeVB=gtk.VBox()
        s.setSizeFrame(chan)
        sizeFrame.add(s.sizeVB)
        return sizeFrame

    def set_dest(s):
        chan = s.channel
        if (chan.dest == 'file') and hasattr(chan,'file'):
            s.fileFrame=gtk.Frame(_('Channel File Options'))
            fileVB=gtk.VBox()
            s.fileWidget=DNS_Path(s,s.rr,chan.file.path,chan.name)
            fileVB.pack_start(s.fileWidget.widget(),False)
            fileVB.pack_start(s.versionFrame(chan),False)
            fileVB.pack_start(s.sizeFrame(chan),False)
            s.fileFrame.add(fileVB)
            s.destHB.pack_start(s.fileFrame,False)
        elif(chan.dest == 'syslog') and hasattr(chan,'syslog'):
            s.logFacilityChoice = gtk.combo_box_new_text()
            c=0
            facility=0
            for f in s.attr.syslog_facilities:
                s.logFacilityChoice.append_text(f)
                if f == chan.syslog:
                    facility=c
                c+=1
            s.logFacilityChoice.set_active(facility)
            s.destHB.pack_start(s.logFacilityChoice,False)
        
    def editChan(s, iter, chan):
        if hasattr(s,'channel') and (s.channel != None):
            debug( 'EDIT CHAN - saving channel:',s.channel.name)
            if (s.channel != chan):
                if not s.saveChan():
                    if hasattr(chan,'new_channel'):
                        s.chanList.remove(iter)
                    return
            else:
                return
        s.channel  = chan
        s.chanIter = iter
        debug( 'EDIT CHAN:',s.channel.name)
        for c in  s.chanVB.get_children():
            s.chanVB.remove(c)
        if chan.name in [ 'default_syslog', 'default_debug', 'default_stderr', 'null' ] :
            frame = gtk.Frame(_('Default Channel') + ' ' + chan.name)
            vb=gtk.VBox()
            destFrame=gtk.Frame(_('Destination'))
            hb = gtk.HBox()
            hb.pack_start(gtk.Label(chan.dest),False)
            hb.pack_start(gtk.Label(str(getattr(chan,chan.dest))))
            destFrame.add(hb)
            vb.pack_start(destFrame,False)
            if hasattr(chan,'severity'):                
                sevFrame=gtk.Frame(_('Severity'))
                hb = gtk.HBox()
                hb.pack_start(gtk.Label(chan.severity),False)
                sevFrame.add(hb)
                vb.pack_start(sevFrame)
            frame.add(vb)
            butHB=gtk.HBox()
            s.chanVB.pack_start(frame)
            s.addChanButton=gtk.Button(stock=gtk.STOCK_ADD)
            s.addChanButton.connect('clicked',s.addChan)
            s.removeChanButton=gtk.Button(stock=gtk.STOCK_REMOVE)
            s.removeChanButton.connect('clicked',s.removeChan)
            butHB.pack_start(s.addChanButton,False)
            butHB.pack_start(s.removeChanButton,False)
            s.chanVB.pack_start(butHB,False)
            s.chanVB.show_all()
            return
        frame=gtk.Frame(_('Channel Name'))
        hb=gtk.HBox()
        s.chanName = gtk.Entry()
        s.chanName.connect('focus_out_event',s.checkChanName)
        s.chanName.set_width_chars(max(len(chan.name),20))
        s.chanName.set_text(chan.name)
        hb.pack_start(s.chanName,False)
        frame.add(hb)
        s.chanVB.pack_start(frame,False)
        destFrame=gtk.Frame(_('Channel Destination'))
        channel_dest=''
        if hasattr(chan,'dest') and (len(chan.dest)>0):
            channel_dest=chan.dest
            debug( 'EDIT CHAN: dest:',channel_dest        )
        s.destHB=gtk.HBox()
        s.destChoice = gtk.combo_box_new_text()
        d=0
        activeDest=0
        for dest in s.attr.destinations:
            s.destChoice.append_text(dest)
            if channel_dest == dest:
                activeDest=d
            d+=1
        s.destChoice.set_active(activeDest)
        s.destChoice.connect('changed',s.changeDest)
        dvb=gtk.VBox()
        dvb.pack_start(s.destChoice,False)
        s.destHB.pack_start(dvb,False)
        s.set_dest()
        destFrame.add(s.destHB)
        s.chanVB.pack_start(destFrame,False)
        sevFrame=gtk.Frame(_('Channel Severity'))
        s.severityChoice=gtk.combo_box_new_text()
        severity='info'
        if hasattr(s.attr,'severity'):
            severity = s.attr.severity
        c=0
        for sev in s.attr.severities:
            s.severityChoice.append_text(sev)
            if sev == severity:
                s.severityChoice.set_active(c)                
            c+=1
        s.severityChoice.connect('changed',s.changeSeverity)
        shb=gtk.HBox()
        shb.pack_start(s.severityChoice,False)
        sevFrame.add(shb)
        s.chanVB.pack_start(sevFrame,False)
        optFrame=gtk.Frame(_('Channel Print Options'))
        optVB=gtk.HBox()
        s.optPrintTime=gtk.CheckButton(label=_('Print Time'))
        if chan.opts.has_key('print-time'):
            if chan.opts['print-time'] == 'no':
                del chan.opts['print-time']
            else:
                s.optPrintTime.set_active(True)
        optVB.pack_start(s.optPrintTime,False)
        s.optPrintSeverity=gtk.CheckButton(label=_('Print Severity'))
        if chan.opts.has_key('print-severity'):
            if chan.opts['print-severity'] == 'no':
                del chan.opts['print-severity']
            else:
                s.optPrintSeverity.set_active(True)
        optVB.pack_start(s.optPrintSeverity,False)
        s.optPrintCategory=gtk.CheckButton(label=_('Print Category'))
        if chan.opts.has_key('print-category'):
            if chan.opts['print-category'] == 'no':
                del chan.opts['print-category']
            else:
                s.optPrintCategory.set_active(True)
        optVB.pack_start(s.optPrintCategory,False)
        optFrame.add(optVB)
        s.chanVB.pack_start(optFrame)
        butHB=gtk.HBox()
        s.addChanButton=gtk.Button(stock=gtk.STOCK_ADD)
        s.addChanButton.connect('clicked',s.addChan)
        s.removeChanButton=gtk.Button(stock=gtk.STOCK_REMOVE)
        s.removeChanButton.connect('clicked',s.removeChan)
        butHB.pack_start(s.addChanButton,False)
        butHB.pack_start(s.removeChanButton,False)
        s.chanVB.pack_start(butHB,False)
        s.chanVB.show_all()
        s.cat_element_selected(s.catTree,0)
        
    def saveCatChan(s, model, path, iter, cat):
        chan = model.get_value(iter,0)
        cat.channels.append(chan)
        debug( 'SAVE CAT:',cat.name,' ADD:', chan        )
    
    def saveCategory(s):
        s.category.channels=[]
        s.catChanList.foreach(s.saveCatChan,s.category)
    
    def add_catChan(s,catChan):
        (iter,chan) = s.selected(s.catAllChanTree)
        if chan == None:
            s.gui.modalError(s.parent,_('Category Channel Add Error'),_('You must select a category channel to add.'))
            return
        s.catChanList.append([s.catAllChanList.get_value(iter,0),s.catChanList.iter_n_children(None)])
        
    def remove_catChan(s,catChan):
        (iter,chan) = s.selected(s.catChanTree)
        if chan == None:
            s.gui.modalError(s.parent,_('Category Channel Remove Error'),_('You must select a category channel to remove.'))
            return
        s.catChanList.remove(iter)

    def select_category(s,w):
        i=s.catCombo.get_active()
        if(i <= -1) or (i > len(s.catsToAdd)):
            debug( 'BAD CATEGORY SELECTION:',i)
        cat = s.catsToAdd[ i ]
        debug( 'SELECT CAT:',i,cat        )
        s.catInfoLabel.set_markup(_('Category')+' <b>' + cat + '</b> '+ _('Description'))
        s.catInfo.set_markup(DNS.Logging.log_categories[cat]['desc'])
        return False

    def add_category(s,w):
        i=s.catCombo.get_active()
        if(i <= -1) or (i > len(s.catsToAdd)):
            s.gui.modalError(s.parent,_('Category Add Error'),_('You must select a category to add.'))
            return
        cat = s.catsToAdd[i]
        debug( 'ADD CAT:',i,cat)
        s.attr.categories.append( DNS.Logging.Category(s.attr,cat,{}) )
        n = s.catList.append([cat, s.attr.categories[-1]])
        cats=s.catKeys.keys()
        for i in range(0,len(cats)):
            s.catCombo.remove_text(i)        
        s.catKeys = {}
        for cat in s.attr.categories:
            s.catKeys[cat.name]=cat
        s.catsToAdd=[]
        catList=DNS.Logging.log_categories.keys()
        catList.sort()
        for cat in catList:
            if not s.catKeys.has_key(cat):
                s.catsToAdd.append(cat)
                s.catCombo.append_text(cat)
        if len(s.catsToAdd):
            s.catCombo.set_active(0)
        s.catTree.set_cursor((len(s.attr.categories)-1,))
        s.editCat(n,s.attr.categories[-1])
        return False
    
    def remove_category(s,w):
        if not hasattr(s,'category'):
            return
        i=-1
        for cat in s.attr.categories:
            i+=1
            if cat == s.category:
                break
        if (i > -1) and (i < len(s.attr.categories)):
            s.attr.categories.pop(i)
        else:
            debug( 'BAD CAT TO REMOVE',i)
            return False
        s.catList.remove(s.catList.get_iter((i,)))
        debug( 'REMOVED ', s.category.name,i,len(s.catsToAdd))
        s.catCombo.get_model().clear()
        s.catKeys = {}
        for cat in s.attr.categories:
            s.catKeys[cat.name]=cat
        s.catsToAdd=[]
        catList=DNS.Logging.log_categories.keys()
        catList.sort()
        i=0
        for cat in catList:
            if not s.catKeys.has_key(cat):
                s.catsToAdd.append(cat)
                s.catCombo.append_text(cat)
            i+=1
        if len(s.catsToAdd)>0:
            s.catCombo.set_active(0)
        debug( ' catsToAdd:',len(s.catsToAdd))
        if s.catList.iter_n_children(None):
            s.catTree.set_cursor((0,))
            s.editCat((0,),s.attr.categories[0])
        return False

    def findChan(s, model, path, iter, chanName):
        chan = model.get_value(iter,1)
        if chan.name == chanName:
            s.selectedChan = chan
            s.selectedChanIter = iter
            
    def selectCatChan(s,iter,catChan):
        debug( 'SELECT CAT CHAN:',catChan)
        s.selectedChan = None
        s.chanList.foreach(s.findChan,catChan)
        if s.selectedChan != None:
            s.editChan(s.selectedChanIter,s.selectedChan)
        
    def editCat(s,iter,cat):
        debug( 'EDIT CAT:',cat.name)
        if hasattr(s,'category'):
            s.saveCategory()
        s.category = cat
        s.catChanList.clear()
        chans={}
        i=0
        for catChan in cat.channels:            
           chans[catChan] = s.catChanList.append([catChan,i])
           i+=1
        if i > 0:
            s.catChanTree.set_cursor((0,))
        s.catAllChanList.clear()
        if not s.attr.channels.has_key(s.channel.name):
            s.saveChan()
        i=0
        for chan in s.attr.channels.keys():
            if not chans.has_key(chan):
                s.catAllChanList.append([chan,i])
            i+=1
        if i > 0:
            s.catAllChanTree.set_cursor((0,))
        s.catChanLabel.set_markup(_('Category')+' <b>' + cat.name + '</b> ')
        s.catInfoLabel.set_markup(_('Category')+' <b>' + cat.name + '</b> '+ _('Description'))
        s.catInfo.set_markup(DNS.Logging.log_categories[cat.name]['desc'])
    
    def selected(s,tree):
        selection = tree.get_selection()
        if( selection != None ):
            model, selected = selection.get_selected()
            if (model != None) and  (selected != None):
                return (selected,) + model.get(selected, 1)
        return (None,None)
    
    def chan_element_activated(s,t,p,c):
        (iter, chan)=s.selected(s.chanTree)
        if( chan != None ):
            s.editChan(iter,chan)
        
    def chan_element_selected(s,t,e):
        (iter, chan)=s.selected(s.chanTree)
        if( chan != None ):
            s.editChan(iter,chan)
    
    def cat_element_activated(s,t,p,c):
        (iter, cat)=s.selected(s.catTree)
        if( cat != None ):
            s.editCat(iter,cat)
            
    def cat_element_selected(s,t,e):
        (iter, cat)=s.selected(s.catTree)
        if( cat != None ):
            s.editCat(iter,cat)

    def catChan_element_activated(s,t,p,c):
        (iter, catChan)=s.selected(s.catChanTree)
        if( iter != None ):
            s.selectCatChan(iter,s.catChanList.get_value(iter,0))
            
    def catChan_element_selected(s,t,e):
        (iter, catChan)=s.selected(s.catChanTree)
        if( iter != None ):
            s.selectCatChan(iter,s.catChanList.get_value(iter,0))

    def catAllChan_element_activated(s,t,p,c):
        (iter, catChan)=s.selected(s.catAllChanTree)
        if( iter != None ):
            s.selectCatChan(iter,s.catAllChanList.get_value(iter,0))
            
    def catAllChan_element_selected(s,t,e):
        (iter, catChan)=s.selected(s.catAllChanTree)
        if( iter != None ):
            s.selectCatChan(iter,s.catAllChanList.get_value(iter,0))
        
    def load(s):
        s.chanList.clear()
        s.catList.clear()
        i=0
        defaults=[]
        for k in s.attr.channels.keys():
            if k not in [ 'default_syslog', 'default_debug', 'default_stderr', 'null']:
                s.chanList.append([k, s.attr.channels[k]])
            else:
                defaults.append([k, s.attr.channels[k]])
            i+=1
        for k in defaults:
            s.chanList.append(k)
        if i > 0:
            s.chanTree.set_cursor((0,))
            s.chan_element_selected(s.chanTree,1)                        
        else:
            s.addChan(1)
            s.chanTree.set_cursor((0,))
        i=0
        for k in s.attr.categories:
            s.catList.append([k.name, k])
            i+=1
        if i > 0:
            s.editCat((0,),s.attr.categories[0])
            s.catTree.set_cursor((0,))

    def save(s):
        if not s.saveChan():
            return False
        s.saveCategory()
        if s.bind.nmdc.__contains__('logging'):
            del s.bind.nmdc['logging']
        s.bind.nmdc['logging']=s.attr.out()
        s.bind.nmdc.update()
        s.bind.zoneTree.load()
        return True
        
    def widget(s):
        s.vb = gtk.VBox()
        s.chanFrame = gtk.Frame(_('Channels'))
        s.chanHB=gtk.HBox()
        s.chanList = gtk.ListStore( gobject.TYPE_STRING, gobject.TYPE_PYOBJECT )
        s.chanTree = gtk.TreeView( model=s.chanList )
        s.chanTree.set_headers_visible(False)
        s.chanTree.connect('row_activated',s.chan_element_activated)
        s.chanTree.connect('button_release_event',s.chan_element_selected)
        s.chanCell = gtk.CellRendererText()
        s.chanColm = gtk.TreeViewColumn( None, s.chanCell, text=0 )
        s.chanTree.append_column(s.chanColm)
        s.chanSW = gtk.ScrolledWindow()
        s.chanSW.set_policy(gtk.POLICY_AUTOMATIC,gtk.POLICY_AUTOMATIC)
        s.chanSW.add_with_viewport(s.chanTree)
        s.chanHB.pack_start(s.chanSW)
        s.chanVB=gtk.VBox()        
        s.chanHB.pack_start(s.chanVB,False)
        s.chanFrame.add(s.chanHB)
        s.vb.pack_start(s.chanFrame)
        s.catFrame = gtk.Frame(_('Categories'))
        s.catHB=gtk.HBox()
        s.catList = gtk.ListStore( gobject.TYPE_STRING, gobject.TYPE_PYOBJECT )
        s.catTree = gtk.TreeView( model=s.catList )
        s.catTree.set_headers_visible(False)
        s.catTree.connect('row_activated',s.cat_element_activated)
        s.catTree.connect('button_release_event',s.cat_element_selected)
        s.catCell = gtk.CellRendererText()
        s.catColm = gtk.TreeViewColumn( None, s.catCell, text=0 )
        s.catTree.append_column(s.catColm)
        s.catSW = gtk.ScrolledWindow()
        s.catSW.set_policy(gtk.POLICY_AUTOMATIC,gtk.POLICY_AUTOMATIC)
        s.catSW.add_with_viewport(s.catTree)
        s.catHB.pack_start(s.catSW)
        s.catVB=gtk.VBox()
        s.catChanListFrame = gtk.Frame(_('Category Channels'))
        s.catChanList = gtk.ListStore( gobject.TYPE_STRING, gobject.TYPE_PYOBJECT )
        s.catChanTree = gtk.TreeView( model=s.catChanList )
        s.catChanTree.set_headers_visible(False)
        s.catChanTree.connect('row_activated',s.catChan_element_activated)
        s.catChanTree.connect('button_release_event',s.catChan_element_selected)
        s.catChanCell = gtk.CellRendererText()
        s.catChanColm = gtk.TreeViewColumn( None, s.catChanCell, text=0 )
        s.catChanTree.append_column(s.catChanColm)
        s.catChanSW = gtk.ScrolledWindow()
        s.catChanSW.set_policy(gtk.POLICY_AUTOMATIC,gtk.POLICY_AUTOMATIC)
        s.catChanSW.add_with_viewport(s.catChanTree)
        s.catChanListFrame.add(s.catChanSW)
        s.catAllChanListFrame=gtk.Frame(_('All Channels'))
        s.catAllChanList = gtk.ListStore( gobject.TYPE_STRING, gobject.TYPE_PYOBJECT )
        s.catAllChanTree = gtk.TreeView( model=s.catAllChanList )
        s.catAllChanTree.set_headers_visible(False)
        s.catAllChanCell = gtk.CellRendererText()
        s.catAllChanColm = gtk.TreeViewColumn( None, s.catAllChanCell, text=0 )
        s.catAllChanTree.append_column(s.catAllChanColm)
        s.catAllChanSW = gtk.ScrolledWindow()
        s.catAllChanSW.set_policy(gtk.POLICY_AUTOMATIC,gtk.POLICY_AUTOMATIC)
        s.catAllChanSW.add_with_viewport(s.catAllChanTree)
        s.catAllChanTree.connect('row_activated',s.catAllChan_element_activated)
        s.catAllChanTree.connect('button_release_event',s.catAllChan_element_selected)        
        caclHB=gtk.HBox()
        caclHB.pack_start(s.catAllChanSW)
        s.catAllChanListFrame.add(caclHB)
        s.catChanFrame=gtk.Frame()
        s.catChanLabel=gtk.Label(_('Category'))
        s.catChanFrame.set_label_widget(s.catChanLabel)
        s.catChanHB=gtk.HBox()
        s.catChanHB.pack_start(s.catChanListFrame)
        s.removeChanButton = gtk.Button(stock=gtk.STOCK_REMOVE)
        s.removeChanButton.connect('clicked',s.remove_catChan)
        rvb=gtk.VBox()
        rhb=gtk.HBox()
        rhb.pack_start(s.removeChanButton,False)
        rvb.pack_start(gtk.Label(''),False)
        rvb.pack_start(rhb,False)
        rvb.pack_start(gtk.Label(''),False)
        s.catChanHB.pack_start(rvb,False)
        s.addChanButton = gtk.Button(stock=gtk.STOCK_ADD)
        s.addChanButton.connect('clicked',s.add_catChan)
        avb=gtk.VBox()
        ahb=gtk.HBox()
        ahb.pack_start(s.addChanButton,False)
        avb.pack_start(gtk.Label(''),False)
        avb.pack_start(ahb,False)
        avb.pack_start(gtk.Label(''),False)
        s.catChanHB.pack_start(avb,False)
        s.catChanHB.pack_start(s.catAllChanListFrame)
        s.catVB.pack_start(s.catChanHB,False)
        s.catInfo=gtk.Label('')
        s.catInfo.set_line_wrap(True)
        s.catInfoFrame=gtk.Frame()
        s.catInfoLabel = gtk.Label(_('Category Description'))
        s.catInfoFrame.set_label_widget(s.catInfoLabel)
        s.catInfoSW=gtk.ScrolledWindow()
        s.catInfoSW.set_policy(gtk.POLICY_AUTOMATIC,gtk.POLICY_AUTOMATIC)
        s.catInfoSW.add_with_viewport(s.catInfo)
        s.catInfoFrame.add(s.catInfoSW)
        s.catVB.pack_start(s.catInfoFrame)
        s.catButHB=gtk.HBox()
        s.catCombo=gtk.combo_box_new_text()
        s.catKeys = {}
        for cat in s.attr.categories:
            s.catKeys[cat.name]=cat
        s.catsToAdd=[]
        catList=DNS.Logging.log_categories.keys()
        catList.sort()
        for cat in catList:
            if not s.catKeys.has_key(cat):
                s.catsToAdd.append(cat)
                s.catCombo.append_text(cat)
        if len(s.catsToAdd):
            s.catCombo.set_active(0)
        s.catCombo.connect('changed',s.select_category)
        s.catButHB.pack_start(s.catCombo,False)        
        s.addCatButton = gtk.Button(stock=gtk.STOCK_ADD)
        s.addCatButton.connect('clicked',s.add_category)
        s.catButHB.pack_start(s.addCatButton,False)
        s.removeCatButton=gtk.Button(stock=gtk.STOCK_REMOVE)
        s.removeCatButton.connect('clicked',s.remove_category)
        s.catButHB.pack_start(s.removeCatButton,False)
        s.catVB.pack_start(s.catButHB,False)
        s.catChanFrame.add(s.catVB)
        s.catHB.pack_start(s.catChanFrame,False)        
        s.catFrame.add(s.catHB)
        s.vb.pack_start(s.catFrame,False)
        s.load()
        return s.vb
    
    def response(s,d,r):
        if r == gtk.RESPONSE_OK:
            if not s.save():
                return False
        s.logDialog.destroy()
        return False
    
    def loggingDialog(s):
        s.logDialog=gtk.Dialog(parent=s.gui.gui, title=_('DNS Logging'),
                            buttons=
                            (gtk.STOCK_OK,     gtk.RESPONSE_OK,
                             gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL
                            )
                           )
        s.logDialog.vbox.pack_start(s.widget())
        s.logDialog.connect('response',s.response)
        s.logDialog.show_all()
        s.parent=s.logDialog

class null_widget(Widget):
    def __init__(s, d, r, a, k):
        Widget.__init__(s, d, r, a, k)

    def widget(s):
        s.label= gtk.Label(str(s.attr))
        return s.label

    def save(s):
        return False

    def value(s):
        return s.label.get_text()

def widget(module, obj):
    if hasattr(obj,'__class__'):
        clas = str(obj.__class__)
        clas = clas.replace('.','_')
        if hasattr(module,clas):
            return getattr(module,clas)
        if( obj.__class__ == int) and (hasattr(module,'int_widget')):
            return module.int_widget
        elif( obj.__class__ == dict) and (hasattr(module,'dict_widget')):
            return module.dict_widget
        elif( obj.__class__ == list) and (hasattr(module,'list_widget')):
            return module.dict_widget
        elif( obj.__class__ == type(None) ) and (hasattr(module,'null_widget')):
            return module.null_widget        
    return Widget
